from __future__ import annotations

import pathlib
import tempfile
import textwrap
from io import BytesIO
from typing import Callable, Iterator

import pytest

from dissect.target.exceptions import RegistryKeyNotFoundError
from dissect.target.filesystem import Filesystem, VirtualFilesystem, VirtualSymlink
from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugin import OSPlugin
from dissect.target.plugins.general import default
from dissect.target.plugins.os.unix._os import UnixPlugin
from dissect.target.plugins.os.unix.bsd.citrix._os import CitrixPlugin
from dissect.target.plugins.os.unix.bsd.osx._os import MacPlugin
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.plugins.os.unix.linux.android._os import AndroidPlugin
from dissect.target.plugins.os.unix.linux.debian._os import DebianPlugin
from dissect.target.plugins.os.unix.linux.redhat._os import RedHatPlugin
from dissect.target.plugins.os.unix.linux.suse._os import SuSEPlugin
from dissect.target.plugins.os.windows import registry
from dissect.target.plugins.os.windows._os import WindowsPlugin
from dissect.target.target import Target
from tests._utils import absolute_path

# Test if the data/ directory is present and if not, as is the case in Python
# source distributions of dissect.target, we give an error
data_dir = absolute_path("_data")
if not pathlib.Path(data_dir).is_dir():
    raise pytest.PytestConfigWarning(
        f"No test data directory {data_dir} found.\n"
        "This can happen when you have downloaded the source distribution\n"
        "of dissect.target from pypi.org. If so, retrieve the test data from\n"
        "the dissect.target GitHub repository at:\n"
        "https://github.com/fox-it/dissect.target"
    )


def make_mock_target(tmp_path: pathlib.Path) -> Iterator[Target]:
    with tempfile.NamedTemporaryFile(dir=tmp_path, prefix="MockTarget-", delete=False) as tmp_file:
        tmp_file.close()
        target = Target()
        target.path = pathlib.Path(tmp_file.name)
        yield target


def make_os_target(
    tmp_path: pathlib.Path,
    os_plugin: type[OSPlugin],
    root_fs: Filesystem | None = None,
    apply_target: bool = True,
) -> Target:
    mock_target = next(make_mock_target(tmp_path))
    mock_target._os_plugin = os_plugin

    if root_fs is not None:
        mock_target.filesystems.add(root_fs)

    if apply_target:
        mock_target.apply()

    return mock_target


@pytest.fixture
def make_mock_targets(request: pytest.FixtureRequest, tmp_path: pathlib.Path) -> Callable[[int], Iterator[Target]]:
    def _make_targets(size: int) -> Iterator[Target]:
        for _ in range(size):
            yield from make_mock_target(tmp_path)

    return _make_targets


@pytest.fixture
def fs_win(tmp_path: pathlib.Path) -> Iterator[VirtualFilesystem]:
    fs = VirtualFilesystem(case_sensitive=False, alt_separator="\\")
    fs.map_dir("windows/system32", tmp_path)
    fs.map_dir("windows/system32/config/", tmp_path)
    yield fs


@pytest.fixture
def fs_unix() -> Iterator[VirtualFilesystem]:
    fs = VirtualFilesystem()
    fs.makedirs("var")
    fs.makedirs("etc")
    yield fs


@pytest.fixture
def fs_linux() -> Iterator[VirtualFilesystem]:
    fs = VirtualFilesystem()
    fs.makedirs("var")
    fs.makedirs("etc")
    fs.makedirs("opt")
    yield fs


@pytest.fixture
def fs_debian() -> Iterator[VirtualFilesystem]:
    fs = VirtualFilesystem()
    fs.makedirs("var")
    fs.makedirs("etc/dpkg")
    fs.makedirs("opt")
    yield fs


@pytest.fixture
def fs_redhat() -> Iterator[VirtualFilesystem]:
    fs = VirtualFilesystem()
    fs.makedirs("var")
    fs.makedirs("etc/sysconfig/network-scripts")
    fs.makedirs("opt")
    yield fs


@pytest.fixture
def fs_suse() -> Iterator[VirtualFilesystem]:
    fs = VirtualFilesystem()
    fs.makedirs("var")
    fs.makedirs("etc/zypp")
    fs.makedirs("opt")
    yield fs


@pytest.fixture
def fs_linux_sys(fs_linux: VirtualFilesystem) -> Iterator[VirtualFilesystem]:
    fs_linux.makedirs("sys")
    yield fs_linux


@pytest.fixture
def fs_linux_proc(fs_linux: VirtualFilesystem) -> Iterator[VirtualFilesystem]:
    fs = fs_linux

    procs = (
        (
            "proc/1",
            VirtualSymlink(fs, "/proc/1/fd/4", "socket:[1337]"),
            "test\x00cmdline\x00",
            "VAR=1",
        ),
        (
            "proc/2",
            VirtualSymlink(fs, "/proc/2/fd/4", "socket:[1338]"),
            "\x00",
            "VAR=1\x00",
        ),
        (
            "proc/3",
            VirtualSymlink(fs, "/proc/3/fd/4", "socket:[1339]"),
            "sshd",
            "VAR=1",
        ),
        (
            "proc/1337",
            VirtualSymlink(fs, "/proc/1337/fd/4", "socket:[1337]"),
            "acquire\x00-p\x00full\x00--proc\x00",
            "VAR=1",
        ),
    )
    stat_files_data = (
        "1 (systemd) S 0 1 1 0 -1 4194560 53787 457726 166 4255 112 260 761 548 20 0 1 0 30 184877056 2658 18446744073709551615 93937510957056 93937511789581 140726499200496 0 0 0 671173123 4096 1260 0 0 0 17 0 0 0 11 0 0 93937512175824 93937512476912 93937519890432 140726499204941 140726499204952 140726499204952 140726499205101 0\n",  # noqa
        "2 (kthread) K 1 1 1 0 -1 4194560 53787 457726 166 4255 112 260 761 548 20 0 1 0 30 184877056 2658 18446744073709551615 93937510957056 93937511789581 140726499200496 0 0 0 671173123 4096 1260 0 0 0 17 0 0 0 11 0 0 93937512175824 93937512476912 93937519890432 140726499204941 140726499204952 140726499204952 140726499205101 0\n",  # noqa
        "3 (sshd) W 1 2 1 0 -1 4194560 53787 457726 166 4255 112 260 761 548 20 0 1 0 30 184877056 2658 18446744073709551615 93937510957056 93937511789581 140726499200496 0 0 0 671173123 4096 1260 0 0 0 17 0 0 0 11 0 0 93937512175824 93937512476912 93937519890432 140726499204941 140726499204952 140726499204952 140726499205101 0\n",  # noqa
        "1337 (acquire) R 3 1 1 0 -1 4194560 53787 457726 166 4255 112 260 761 548 20 0 1 0 30 184877056 2658 18446744073709551615 93937510957056 93937511789581 140726499200496 0 0 0 671173123 4096 1260 0 0 0 17 0 0 0 11 0 0 93937512175824 93937512476912 93937519890432 140726499204941 140726499204952 140726499204952 140726499205101 0\n",  # noqa
    )

    for idx, proc in enumerate(procs):
        dir, fd, cmdline, environ = proc
        fs.makedirs(dir)
        fs.map_file_entry(fd.path, fd)

        fs.map_file_fh(dir + "/stat", BytesIO(stat_files_data[idx].encode()))
        fs.map_file_fh(dir + "/cmdline", BytesIO(cmdline.encode()))
        fs.map_file_fh(dir + "/environ", BytesIO(environ.encode()))

    # symlink acquire process to self
    fs.link("/proc/1337", "/proc/self")

    # boottime and uptime are needed for for time tests
    fs.map_file_fh("/proc/uptime", BytesIO(b"134368.27 132695.52\n"))
    fs.map_file_fh("/proc/stat", BytesIO(b"btime 1680559854"))

    yield fs


@pytest.fixture
def fs_linux_proc_sockets(fs_linux_proc: VirtualFilesystem) -> Iterator[VirtualFilesystem]:
    fs = fs_linux_proc

    for filename in ("unix", "packet", "raw6", "raw", "udp6", "udp", "tcp6", "tcp"):
        fs.map_file(f"/proc/net/{filename}", absolute_path(f"_data/plugins/os/unix/linux/proc/net/{filename}"))

    yield fs


@pytest.fixture
def fs_osx() -> Iterator[VirtualFilesystem]:
    fs = VirtualFilesystem()
    fs.makedirs("Applications")
    fs.makedirs("Library")
    yield fs


@pytest.fixture
def fs_bsd() -> Iterator[VirtualFilesystem]:
    fs = VirtualFilesystem()
    fs.map_file("/bin/freebsd-version", absolute_path("_data/plugins/os/unix/bsd/freebsd/freebsd-freebsd-version"))
    yield fs


@pytest.fixture
def fs_android() -> Iterator[VirtualFilesystem]:
    fs = VirtualFilesystem()
    fs.map_file("/build.prop", absolute_path("_data/plugins/os/unix/linux/android/build.prop"))
    yield fs


@pytest.fixture
def hive_hklm() -> Iterator[VirtualHive]:
    hive = VirtualHive()

    # set current control set to ControlSet001 and mock it
    change_controlset(hive, 1)

    yield hive


def change_controlset(hive: VirtualHive, num: int) -> None:
    """Update the current control set of the given HKLM hive."""

    if not isinstance(num, int) or num > 999 or num < 1:
        raise ValueError("ControlSet integer must be between 1 and 999")

    controlset_key = f"SYSTEM\\ControlSet{num:>03}"
    hive.map_key(controlset_key, VirtualKey(hive, controlset_key))

    select_key = "SYSTEM\\Select"
    hive.map_key(select_key, VirtualKey(hive, select_key))
    hive.map_value(select_key, "Current", VirtualValue(hive, "Current", num))


@pytest.fixture
def hive_hku() -> Iterator[VirtualHive]:
    hive = VirtualHive()

    yield hive


@pytest.fixture
def target_bare(tmp_path: pathlib.Path) -> Iterator[Target]:
    # A target without any associated OSPlugin
    yield from make_mock_target(tmp_path)


@pytest.fixture
def target_default(tmp_path: pathlib.Path) -> Iterator[Target]:
    yield make_os_target(tmp_path, default.DefaultPlugin)


@pytest.fixture
def target_win(tmp_path: pathlib.Path, hive_hklm: VirtualHive, fs_win: Filesystem) -> Iterator[Target]:
    mock_target = make_os_target(tmp_path, WindowsPlugin, root_fs=fs_win, apply_target=False)

    mock_target.add_plugin(registry.RegistryPlugin, check_compatible=False)
    mock_target.registry.add_hive(
        "HKEY_LOCAL_MACHINE",
        "HKEY_LOCAL_MACHINE",
        hive_hklm,
        TargetPath(mock_target.fs, ""),
    )
    mock_target.fs.mount("c:", fs_win)

    mock_target.apply()

    yield mock_target


@pytest.fixture
def target_unix(tmp_path: pathlib.Path, fs_unix: Filesystem) -> Iterator[Target]:
    yield make_os_target(tmp_path, UnixPlugin, root_fs=fs_unix)


@pytest.fixture
def target_linux(tmp_path: pathlib.Path, fs_linux: Filesystem) -> Iterator[Target]:
    yield make_os_target(tmp_path, LinuxPlugin, root_fs=fs_linux)


@pytest.fixture
def target_debian(tmp_path: pathlib.Path, fs_debian: Filesystem) -> Iterator[Target]:
    yield make_os_target(tmp_path, DebianPlugin, root_fs=fs_debian)


@pytest.fixture
def target_redhat(tmp_path: pathlib.Path, fs_redhat: Filesystem) -> Iterator[Target]:
    yield make_os_target(tmp_path, RedHatPlugin, root_fs=fs_redhat)


@pytest.fixture
def target_suse(tmp_path: pathlib.Path, fs_suse: Filesystem) -> Iterator[Target]:
    yield make_os_target(tmp_path, SuSEPlugin, root_fs=fs_suse)


@pytest.fixture
def target_osx(tmp_path: pathlib.Path, fs_osx: Filesystem) -> Iterator[Target]:
    mock_target = make_os_target(tmp_path, MacPlugin, root_fs=fs_osx)

    version = absolute_path("_data/plugins/os/unix/bsd/osx/_os/SystemVersion.plist")
    fs_osx.map_file("/System/Library/CoreServices/SystemVersion.plist", version)

    system = absolute_path("_data/plugins/os/unix/bsd/osx/_os/preferences.plist")
    fs_osx.map_file("/Library/Preferences/SystemConfiguration/preferences.plist", system)

    yield mock_target


@pytest.fixture
def target_citrix(tmp_path: pathlib.Path, fs_bsd: VirtualFilesystem) -> Target:
    mock_target = next(make_mock_target(tmp_path))
    mock_target._os_plugin = CitrixPlugin

    mock_target.filesystems.add(fs_bsd)

    var_filesystem = VirtualFilesystem()
    var_filesystem.makedirs("/netscaler")
    var_filesystem.makedirs("/log")
    mock_target.filesystems.add(var_filesystem)

    flash_filesystem = VirtualFilesystem()
    flash_filesystem.map_dir("/", absolute_path("_data/plugins/os/unix/bsd/citrix/_os/flash"))
    mock_target.filesystems.add(flash_filesystem)

    mock_target.apply()
    yield mock_target


@pytest.fixture
def target_android(tmp_path: pathlib.Path, fs_android: Filesystem) -> Iterator[Target]:
    yield make_os_target(tmp_path, AndroidPlugin, root_fs=fs_android)


def add_win_user(hive_hklm: VirtualHive, hive_hku: VirtualHive, target_win: Target, sid: str, home: str) -> None:
    """add a user to the provided windows target"""

    profile_list_key_name = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList"
    try:
        profile_list_key = hive_hklm.key(profile_list_key_name)
    except RegistryKeyNotFoundError:
        profile_list_key = VirtualKey(hive_hklm, profile_list_key_name)

    try:
        hive_hklm.key(f"{profile_list_key_name}\\{sid}")
        raise ValueError(f"User with SID {sid} already exists!")
    except RegistryKeyNotFoundError:
        pass

    profile_key = VirtualKey(hive_hklm, f"{profile_list_key_name}\\{sid}")
    profile_key.add_value("ProfileImagePath", VirtualValue(hive_hklm, "ProfileImagePath", home))
    profile_list_key.add_subkey(sid, profile_key)

    hive_hklm.map_key(profile_list_key_name, profile_list_key)

    if sid != "S-1-5-18":
        target_win.registry.add_hive("HKEY_USERS", f"HKEY_USERS\\{sid}", hive_hku, TargetPath(target_win.fs, ""))


@pytest.fixture
def target_win_users(hive_hklm: VirtualHive, hive_hku: VirtualHive, target_win: Target) -> Iterator[Target]:
    """minimal windows target with a SYSTEM user and local user John"""
    add_win_user(hive_hklm, hive_hku, target_win, sid="S-1-5-18", home="%systemroot%\\system32\\config\\systemprofile")
    add_win_user(
        hive_hklm, hive_hku, target_win, sid="S-1-5-21-3263113198-3007035898-945866154-1002", home="C:\\Users\\John"
    )
    yield target_win


@pytest.fixture
def target_win_tzinfo(hive_hklm: VirtualHive, target_win: Target) -> Iterator[Target]:
    tz_info_path = "SYSTEM\\ControlSet001\\Control\\TimeZoneInformation"
    tz_info = VirtualKey(hive_hklm, tz_info_path)
    tz_info.add_value("TimeZoneKeyName", "Easter Island Standard Time")

    eu_tz_data_path = "Software\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\W. Europe Standard Time"
    eu_tz_data = VirtualKey(hive_hklm, eu_tz_data_path)
    eu_tz_data.add_value("Display", "(UTC+01:00) Amsterdam, Berlin, Bern, Rome, Stockholm, Vienna")
    eu_tz_data.add_value("Dlt", "W. Europe Daylight Time")
    eu_tz_data.add_value("Std", "W. Europe Standard Time")
    eu_tzi = bytes.fromhex("c4ffffff00000000c4ffffff00000a0000000500030000000000000000000300000005000200000000000000")
    eu_tz_data.add_value("TZI", eu_tzi)

    east_tz_data_path = "Software\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\Easter Island Standard Time"
    east_tz_data = VirtualKey(hive_hklm, east_tz_data_path)
    east_tz_data.add_value("Display", "(UTC-06:00) Easter Island")
    east_tz_data.add_value("Dlt", "Easter Island Daylight Time")
    east_tz_data.add_value("Std", "Easter Island Standard Time")
    east_tzi = bytes.fromhex("6801000000000000c4ffffff0000040006000100160000000000000000000900060001001600000000000000")
    east_tz_data.add_value("TZI", east_tzi)

    dynamic_dst = VirtualKey(hive_hklm, east_tz_data_path + "\\Dynamic DST")
    dynamic_dst.add_value("FirstEntry", 2019)
    dynamic_dst.add_value("LastEntry", 2019)
    tzi_2019 = bytes.fromhex("6801000000000000c4ffffff0000040006000100160000000000000000000900060001001600000000000000")
    dynamic_dst.add_value("2019", tzi_2019)
    east_tz_data.add_subkey("Dynamic DST", dynamic_dst)

    hive_hklm.map_key(tz_info_path, tz_info)
    hive_hklm.map_key(eu_tz_data_path, eu_tz_data)
    hive_hklm.map_key(east_tz_data_path, east_tz_data)

    yield target_win


@pytest.fixture
def target_win_tzinfo_legacy(hive_hklm: VirtualHive, target_win: Target) -> Iterator[Target]:
    tz_info_path = "SYSTEM\\ControlSet001\\Control\\TimeZoneInformation"
    tz_info = VirtualKey(hive_hklm, tz_info_path)
    tz_info.add_value("StandardName", "Paaseiland")

    east_tz_data_path = "Software\\Microsoft\\Windows NT\\CurrentVersion\\Time Zones\\Easter Island Standard Time"
    east_tz_data = VirtualKey(hive_hklm, east_tz_data_path)
    east_tz_data.add_value("Display", "(UTC-06:00) Easter Island")
    east_tz_data.add_value("Dlt", "Easter Island Daylight Time")
    east_tz_data.add_value("Std", "Paaseiland")
    east_tzi = bytes.fromhex("6801000000000000c4ffffff0000040006000100160000000000000000000900060001001600000000000000")
    east_tz_data.add_value("TZI", east_tzi)

    hive_hklm.map_key(tz_info_path, tz_info)
    hive_hklm.map_key(east_tz_data_path, east_tz_data)

    yield target_win


@pytest.fixture
def target_unix_users(target_unix: Target, fs_unix: Filesystem) -> Iterator[Target]:
    passwd = """
    root:x:0:0:root:/root:/bin/bash
    user:x:1000:1000:user:/home/user:/bin/bash
    """
    fs_unix.map_file_fh("/etc/passwd", BytesIO(textwrap.dedent(passwd).encode()))
    yield target_unix


@pytest.fixture
def target_linux_users(target_linux: Target, fs_linux: VirtualFilesystem) -> Iterator[Target]:
    passwd = """
    root:x:0:0:root:/root:/bin/bash
    user:x:1000:1000:user:/home/user:/bin/bash
    """
    fs_linux.map_file_fh("/etc/passwd", BytesIO(textwrap.dedent(passwd).encode()))
    yield target_linux


@pytest.fixture
def target_osx_users(target_osx: Target, fs_osx: VirtualFilesystem) -> Iterator[Target]:
    dissect = absolute_path("_data/plugins/os/unix/bsd/osx/_os/dissect.plist")
    fs_osx.map_file("/var/db/dslocal/nodes/Default/users/_dissect.plist", dissect)

    test = absolute_path("_data/plugins/os/unix/bsd/osx/_os/test.plist")
    fs_osx.map_file("/var/db/dslocal/nodes/Default/users/_test.plist", test)

    fs_osx.makedirs("/Users/dissect")

    yield target_osx


@pytest.fixture
def fs_docker() -> Iterator[TarFilesystem]:
    docker_tar = pathlib.Path(absolute_path("_data/plugins/apps/container/docker/docker.tgz"))
    fh = docker_tar.open("rb")
    docker_fs = TarFilesystem(fh)
    yield docker_fs


@pytest.fixture
def target_linux_docker(tmp_path: pathlib.Path, fs_docker: TarFilesystem) -> Iterator[Target]:
    mock_target = next(make_mock_target(tmp_path))
    mock_target._os_plugin = LinuxPlugin

    mock_target.filesystems.add(fs_docker)
    mock_target.fs.mount("/", fs_docker)
    mock_target.apply()
    yield mock_target



import os
from datetime import datetime
from io import BytesIO
from pathlib import Path

from dissect.ntfs.secure import ACL, SecurityDescriptor
from flow.record.fieldtypes import command
from flow.record.fieldtypes import datetime as dt

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.helpers.regutil import VirtualHive, VirtualKey
from dissect.target.plugins.os.windows import defender
from dissect.target.target import Target
from tests._utils import absolute_path


def test_defender_evtx_logs(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    # map default log location to pass EvtxPlugin's compatibility check
    fs_win.map_dir("windows/system32/winevt/logs", tmp_path)

    log_file = absolute_path("_data/plugins/os/windows/defender/operational.evtx")
    fs_win.map_file("windows/system32/winevt/logs/Microsoft-Windows-Windows Defender%4Operational.evtx", log_file)

    target_win.add_plugin(defender.MicrosoftDefenderPlugin)

    records = list(target_win.defender.evtx())

    assert len(records) == 9

    # verify that all records have unique EventIDs
    assert all(r.ts is not None for r in records)
    assert len(list(r.EventID for r in records)) == 9
    assert {r.Provider_Name for r in records} == {"Microsoft-Windows-Windows Defender"}
    assert {r.Channel for r in records} == {"Microsoft-Windows-Windows Defender/Operational"}
    # Both informational records (no threat name) and detections are present
    assert {r.Threat_Name for r in records} == {None, "TrojanDropper:PowerShell/PowerSploit.S!MSR"}


def test_defender_quarantine_entries(target_win: Target, fs_win: VirtualFilesystem) -> None:
    quarantine_dir = absolute_path("_data/plugins/os/windows/defender/quarantine")

    fs_win.map_dir("programdata/microsoft/windows defender/quarantine", quarantine_dir)

    target_win.add_plugin(defender.MicrosoftDefenderPlugin)

    records = list(target_win.defender.quarantine())

    assert len(records) == 1

    # Test whether the quarantining of a Mimikatz binary is properly parsed.
    mimikatz_record = records[0]
    detection_date = datetime.strptime("2022-12-02", "%Y-%m-%d").date()

    assert mimikatz_record.detection_type == "file"
    assert mimikatz_record.detection_name == "HackTool:Win64/Mikatz!dha"
    assert mimikatz_record.detection_path == "C:\\Users\\user\\Downloads\\mimikatz\\mimilib.dll"

    assert mimikatz_record.ts.date() == detection_date
    assert mimikatz_record.creation_time.date() == detection_date
    assert mimikatz_record.last_write_time.date() == detection_date
    assert mimikatz_record.last_accessed_time.date() == detection_date


def test_defender_quarantine_recovery(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    # Map the quarantine folder from our test data
    quarantine_dir = absolute_path("_data/plugins/os/windows/defender/quarantine")
    fs_win.map_dir("programdata/microsoft/windows defender/quarantine", quarantine_dir)

    # Create a directory to recover to
    recovery_dst = tmp_path.joinpath("recovery")
    recovery_dst.mkdir()

    # Recover
    target_win.add_plugin(defender.MicrosoftDefenderPlugin)
    target_win.defender.recover(output_dir=recovery_dst)

    # Set up variables to indicate what we expect to find
    payload_filename = "A6C8322B8A19AEED96EFBD045206966DA4C9619D"
    security_descriptor_filename = "A6C8322B8A19AEED96EFBD045206966DA4C9619D.security_descriptor"
    zone_identifier_filename = "A6C8322B8A19AEED96EFBD045206966DA4C9619D.ZoneIdentifierDATA"
    expected_zone_identifier_content = (
        b"[ZoneTransfer]\r\nZoneId=3\r\nReferrerUrl=C:\\Users\\user\\Downloads\\mimikatz_trunk.zip\r\n"
    )
    expected_owner = "S-1-5-21-2614236324-1336345114-3023566343-1000"
    expected_group = "S-1-5-21-2614236324-1336345114-3023566343-513"

    expected_files = [payload_filename, security_descriptor_filename, zone_identifier_filename]
    expected_files.sort()

    directory_content = os.listdir(recovery_dst)
    directory_content.sort()
    assert expected_files == directory_content
    # Replaced the mimikatz payload with `DUMMY_PAYLOAD` to avoid defender collecting it
    assert recovery_dst.joinpath(payload_filename).read_bytes() == b"DUMMY_PAYLOAD"

    # Verify that the security descriptors are valid security descriptors
    with open(recovery_dst.joinpath(security_descriptor_filename), "rb") as descriptor_file:
        descriptor_buf = descriptor_file.read()
        descriptor = SecurityDescriptor(BytesIO(descriptor_buf))

        assert isinstance(descriptor.dacl, ACL)
        assert isinstance(descriptor.sacl, ACL)

        assert descriptor.owner == expected_owner
        assert descriptor.group == expected_group

    # Verify valid zone identifier for mimikatz file
    assert recovery_dst.joinpath(zone_identifier_filename).read_bytes() == expected_zone_identifier_content


def test_defender_exclusions(target_win: Target, hive_hklm: VirtualHive) -> None:
    # https://learn.microsoft.com/en-us/exchange/antispam-and-antimalware/windows-antivirus-software?view=exchserver-2019
    exclusions_example = {
        "Extensions": [".config", ".log", ".cfg"],
        "Processes": ["UmService.exe", "UmWorkerProcess.exe"],
        "Paths": ["C:\\System32\\Cluster\\"],
        "IpAddresses": [],
        "TemporaryPaths": [],
    }
    # Recreate the 'Exclusions' registry key based on the example dict
    exclusions_key = VirtualKey(hive_hklm, "Software\\Microsoft\\Windows Defender\\Exclusions")
    for exclusion_type, exclusions in exclusions_example.items():
        exclusion_type_key = VirtualKey(hive_hklm, exclusions_key.path + f"\\{exclusion_type}")
        for exclusion in exclusions:
            exclusion_type_key.add_value(exclusion, 0)
        exclusions_key.add_subkey(exclusion_type, exclusion_type_key)

    hive_hklm.map_key(exclusions_key.path, exclusions_key)
    target_win.add_plugin(defender.MicrosoftDefenderPlugin)

    exclusion_records = list(target_win.defender.exclusions())

    # If an exclusion type does not have any exclusions, no records are returned for that type.
    ip_address_exclusions = [exclusion for exclusion in exclusion_records if exclusion.type == "IpAddresses"]
    temporary_path_exclusions = [exclusion for exclusion in exclusion_records if exclusion.type == "TemporaryPaths"]
    assert len(ip_address_exclusions) == 0
    assert len(temporary_path_exclusions) == 0

    extension_exclusions = [exclusion for exclusion in exclusion_records if exclusion.type == "Extensions"]
    assert len(extension_exclusions) == 3

    assert any(exclusion.value == ".config" for exclusion in extension_exclusions)
    assert any(exclusion.value == ".log" for exclusion in extension_exclusions)
    assert any(exclusion.value == ".cfg" for exclusion in extension_exclusions)

    process_exclusions = [exclusion for exclusion in exclusion_records if exclusion.type == "Processes"]
    assert len(process_exclusions) == 2
    assert any(exclusion.value == "UmService.exe" for exclusion in process_exclusions)
    assert any(exclusion.value == "UmWorkerProcess.exe" for exclusion in process_exclusions)

    path_exclusions = [exclusion for exclusion in exclusion_records if exclusion.type == "Paths"]
    assert any(exclusion.value == "C:\\System32\\Cluster\\" for exclusion in path_exclusions)
    assert len(path_exclusions) == 1

    assert len(exclusion_records) == 6


def _mplog_records(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path, log_filename: str):
    # map default log location to pass EvtxPlugin's compatibility check
    fs_win.map_dir("windows/system32/winevt/logs", tmp_path)

    log_file = absolute_path(f"_data/plugins/os/windows/defender/mplog/{log_filename}.log")
    fs_win.map_file("ProgramData/Microsoft/Windows Defender/Support/MPLog-20240101-094808.log", log_file)

    target_win.add_plugin(defender.MicrosoftDefenderPlugin)
    records = list(target_win.defender.mplog())
    return records


def test_defender_mplogs_rtp(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    record = _mplog_records(target_win, fs_win, tmp_path, "rtp").pop()
    assert record.source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert record.ts == dt("2021-05-04 09:54:06+00:00")
    assert record.last_perf == dt("2021-05-04 09:54:06+00:00")
    assert record.first_rtp_scan == dt("2021-05-04 09:54:06+00:00")
    assert record.plugin_states == "AV:2  AS:2  RTP:2  OA:2  BM:2"
    assert record.process_exclusions == ["C:\\ReportingServicesService.exe", "C:\\ReportingServicesService.exe"]
    assert sorted(record.path_exclusions) == sorted(
        ["C:\\Windows\\Security\\Database\\*.chk", "%windidr%\\SoftwareDistribution\\Datastore\\*.*"]
    )
    assert sorted(record.ext_exclusions) == sorted([".DBF", ".NDF", ".RAR", ".XML", ".IDX", ".BAK", ".PDF", ".BKP"])


def test_defender_mplogs_resource_scan(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    record = _mplog_records(target_win, fs_win, tmp_path, "resourcescan").pop()
    assert record.source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert record.ts == dt("2023-01-01 00:00:00+00:00")
    assert record.scan_id == "{1A2B3C4D-5E6F-7A8B-9C0D-1E2F3A4B5C6D}"
    assert record.scan_source == 4
    assert record.end_time == dt("2023-01-01 00:01:00+00:00")
    assert record.resource_schema == "webfile"
    assert (
        record.resource_path
        == "C:\\Users\\user\\Downloads\\file.rar|https:/example.com/download?id=12345|browser_broker.exe"
    )
    assert record.result_count == 2
    assert sorted(record.threats) == sorted(["Worm:VBS/Jenxcus.DN", "HackTool:MSIL/Mimikatz!MSR"])
    assert sorted(record.resources) == sorted(
        [
            "".join(
                [
                    "C:\\ProgramData\\App\\Scans\\FilesStash\\9F8E7D6C-5A4B-3C2D-1E0F-A1B2C3D4E5F6_1d23456789abcdef|",
                    "C:\\Users\\user\\Downloads\\file.rar",
                ]
            ),
            "C:\\Users\\user\\Downloads\\file.rar->file.WsF",
            "C:\\ProgramData\\App\\Scans\\FilesStash\\9F8E7D6C-5A4B-3C2D-1E0F-A1B2C3D4E5F6_1d23456789abcdef",
            "C:\\Users\\user\\Downloads\\file.rar|https:/example.com/download?id=67890|browser_broker.exe",
            "".join(
                [
                    "C:\\ProgramData\\App\\Scans\\FilesStash\\9F8E7D6C-5A4B-3C2D-1E0F-A1B2C3D4E5F6_1d23456789abcdef|",
                    "https:/example.com/download?id=12345|browser_broker.exe",
                ]
            ),
            "C:\\Users\\user\\Downloads\\file.rar",
            "C:\\Users\\user\\Videos\\binary.exe",
        ]
    )


def test_defender_mplogs_threat_actions(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    record = _mplog_records(target_win, fs_win, tmp_path, "threatactions").pop()
    assert record.source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert record.ts == dt("2017-12-25 16:30:45+00:00")
    assert record.threats == ["Worm:Win32/RandomName.X"]
    for item in [
        "\\\\?\\C:\\OS\\Tasks\\RandomTask1.job",
        "\\\\?\\C:\\OS\\system32\\randomfile.tmp->(UPX)",
        "\\\\?\\C:\\OS\\Tasks\\RandomTask1.job",
        "\\\\?\\C:\\OS\\system32\\randomfile.tmp",
        "C:\\OS\\Tasks\\RandomTask1.job",
        "\\\\?\\C:\\OS\\Tasks\\RandomTask1.job",
        "C:\\OS\\system32\\randomfile.tmp->(UPX)",
        "\\\\?\\C:\\OS\\system32\\randomfile.tmp->(UPX)",
        "\\\\?\\C:\\OS\\Tasks\\RandomTask1.job",
    ]:
        assert item in list(record.resources)

    assert record.actions == ["quarantine"]


def test_defender_mplogs_bmtelemetry(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    record = _mplog_records(target_win, fs_win, tmp_path, "bmtelemetry").pop()
    assert record.source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert record.ts == dt("2024-07-15 11:45:22+00:00")
    assert record.guid == "{1D3E4F07-89AB-45C2-923D-E5F6789A1B2C}"
    assert record.signature_id == 123456789012345
    assert record.sigsha == "abcd1234ef567890abcd1234ef567890abcd1234"
    assert record.threat_level == 0
    assert record.process_id == 8453
    assert record.process_creation_time == 153846236598765432
    assert record.image_path == "C:\\OS\\System32\\servicehost.exe"
    assert record.taint_info == "Friendly: Y; Reason: ; Modules: ; Parents: "
    assert record.operations == "None"


def test_defender_mplogs_lines(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    records = _mplog_records(target_win, fs_win, tmp_path, "lines")
    assert len(records) == 10

    # Process Image
    assert records[0].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[0].ts == dt("2024-07-13 14:42:19.659000+00:00")
    assert records[0].process_image_name == "randomapp.exe"
    assert records[0].pid == 5832
    assert records[0].total_time == 1398
    assert records[0].count == 22
    assert records[0].max_time == 398
    assert records[0].max_time_file == "\\Device\\HarddiskVolume2\\Users\\user123\\AppData\\Local\\Temp\\TEMP001.tmp"
    assert records[0].estimated_impact == 4

    # Lowfi
    assert records[1].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[1].ts == dt("2023-01-20 08:45:40.321000+00:00")
    assert records[1].lowfi == command(
        "".join(
            [
                "C:\\OS\\System32\\cfg.exe(reg add HKLM\\SYSTEM\\",
                "OtherControlSet\\Control\\SecurityOptions\\SecurityModule /v RandomFlag /t REG_DWORD /d 0 /f)",
            ]
        ),
    )

    # Detection Add
    assert records[2].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[2].ts == dt("2023-01-27 15:33:07.698000+00:00")
    assert records[2].detection == "".join(
        [
            "HackTool:MSIL/RndGen!MD5 file:C:\\Users\\user987\\Documents\\executable.exe ",
            "PropBag [length: 0, data: (null)]",
        ]
    )

    # Threat
    assert records[3].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[3].ts == dt("2023-01-27 15:33:07.698000+00:00")
    assert records[3].threat == command("C:\\Users\\user987\\Documents\\executable.exe")

    # Detection event
    assert records[4].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[4].ts == dt("2023-01-27 15:33:07.698000+00:00")
    assert records[4].threat_type == "MSIL/RndGen!MD5"
    assert records[4].command == command("C:\\Users\\user987\\Documents\\executable.exe")

    # Exclusion
    assert records[5].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[5].ts == dt("2024-08-17 17:35:22.614000+00:00")
    assert records[5].full_path_with_drive_letter == "C:\\example.txt"
    assert records[5].full_path_with_device_path == "example.txt"

    # Mini-filter unsuccesful scan
    assert records[6].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[6].path == "".join(
        [
            "\\Device\\HarddiskVolume2\\Users\\userdefault\\AppData\\Local\\",
            "Packages\\MicrosoftBrowser.Default_cw5n8h2txyuma\\LocalState\\",
            "ANWebView\\Default\\Popular URLs.",
        ]
    )
    assert records[6].ts == dt("2024-07-13 14:38:15.272000+00:00")
    assert records[6].process == "(unknown)"
    assert records[6].status == "0xc000004a"
    assert records[6].state == "0"
    assert records[6].scan_request == "#16891"
    assert records[6].file_id == "0x200000002c4b5"
    assert records[6].reason == "OnAccess"
    assert records[6].io_status_block_for_new_file == "0x3"
    assert records[6].desired_access == "0x0"
    assert records[6].file_attributes == "0x20"
    assert records[6].scan_attributes == "0x10"
    assert records[6].access_state_flags == "0x802"
    assert records[6].backing_file_info == "0x0, 0x0, 0x0:0\\0x0:0"

    # Mini-filter blocked file
    assert records[7].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[7].ts == dt("2024-07-13 14:38:15.272000+00:00")
    assert records[7].blocked_file == "".join(
        [
            "\\Device\\HarddiskVolume2\\Users\\userdefault\\AppData\\Local\\Packages\\",
            "MicrosoftBrowser.Default_cw5n8h2txyuma\\LocalState\\ANWebView\\Default\\Popular URLs.",
        ]
    )
    assert records[7].process == "(unknown)"
    assert records[7].status == "0xc000004a"
    assert records[7].state == "0"
    assert records[7].scan_request == "#16891"
    assert records[7].file_id == "0x200000002c4b5"
    assert records[7].reason == "OnAccess"
    assert records[7].io_status_block_for_new_file == "0x3"
    assert records[7].desired_access == "0x0"
    assert records[7].file_attributes == "0x20"
    assert records[7].scan_attributes == "0x10"
    assert records[7].access_state_flags == "0x802"
    assert records[7].backing_file_info == "0x0, 0x0, 0x0:0\\0x0:0"

    # EMS
    assert records[8].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[8].ts == dt("2024-09-05 10:21:39.417000+00:00")
    assert records[8].process == "sysproc"
    assert records[8].pid == 2820
    assert records[8].sigseq == "0x1"
    assert records[8].send_memory_scan_report == 1
    assert records[8].source == 4

    # Original Filename
    assert records[9].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[9].ts == dt("2024-09-03 18:12:05.364000+00:00")
    assert records[9].original_file_name == "RandomData0123_static.dll"
    assert records[9].full_path == "".join(
        [
            "c:\\os\\winsxs\\x64_default-app-service_31bf3856ad364e35_10.0.29999.9999_none_fakef123456789a\\",
            "randomdata0123.dll",
        ]
    )
    assert records[9].hr == "0x1"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import itertools
import sys
from collections import deque
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Iterable, Iterator, Optional

import structlog
from flow.record import Record

from dissect.target import Target
from dissect.target.exceptions import PluginError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.tools.dump.state import (
    DumpState,
    create_state,
    load_state,
    persisted_state,
)
from dissect.target.tools.dump.utils import (
    Compression,
    Serialization,
    cached_sink_writers,
)
from dissect.target.tools.utils import (
    PluginFunction,
    configure_generic_arguments,
    execute_function_on_target,
    find_and_filter_plugins,
    process_generic_arguments,
)

log = structlog.get_logger("dissect.target.tools.dump.run")


@dataclass
class RecordStreamElement:
    target: Target
    func: str
    record: Record
    end_pos: Optional[int] = None
    sink_path: Optional[Path] = None


def get_targets(targets: list[str]) -> Iterator[Target]:
    """Return a generator with `Target` objects for provided paths"""
    for target in Target.open_all(targets):
        yield target


def execute_function(target: Target, function: PluginFunction) -> TargetRecordDescriptor:
    """
    Execute function `function` on provided target `target` and return a generator
    with the records produced.

    Only output type `record` is supported for plugin functions.
    """

    local_log = log.bind(func=function, target=target)
    local_log.debug("Function execution")

    try:
        output_type, target_attr, _ = execute_function_on_target(target, function)
    except UnsupportedPluginError:
        local_log.error("Function is not supported for target", exc_info=True)
        return
    except PluginError:
        local_log.error("Plugin error while executing function for target", exc_info=True)
        return

    if output_type != "record":
        local_log.warn("Output format is not supported", output=output_type)
        return

    # no support for function-specific arguments
    result = target_attr() if callable(target_attr) else target_attr

    try:
        for value in result:
            yield value
    except Exception:
        local_log.error("Error while executing function", exc_info=True)
        return


def produce_target_func_pairs(
    targets: Iterable[Target],
    functions: str,
    state: DumpState,
) -> Iterator[tuple[Target, PluginFunction]]:
    """
    Return a generator with target and function pairs for execution.

    Target and function pairs that correspond to finished sinks in provided state `state` are skipped.
    """
    pairs_to_skip = set()
    if state:
        pairs_to_skip.update((str(sink.target_path), sink.func) for sink in state.finished_sinks)

    for target in targets:
        for func_def in find_and_filter_plugins(target, functions):
            if state and (target.path, func_def.name) in pairs_to_skip:
                log.info(
                    "Skipping target/func pair since its marked as done in provided state",
                    target=target.path,
                    func=func_def.name,
                    state=state.path,
                )
                continue
            yield (target, func_def)
            state.mark_as_finished(target, func_def.name)


def execute_functions(target_func_stream: Iterable[tuple[Target, str]]) -> Iterable[RecordStreamElement]:
    """
    Execute a function on a target for target / function pairs in the stream.

    Returns a generator of `RecordStreamElement` objects.
    """
    for target, func in target_func_stream:
        for record in execute_function(target, func):
            yield RecordStreamElement(target=target, func=func, record=record)


def log_progress(stream: Iterable[Any], step_size: int = 1000) -> Iterable[Any]:
    """
    Log a number of items that went though the generator stream
    after every N element (N is configured in `step_size`).
    """
    i = 0
    targets = set()
    sinks = set()
    for i, element in enumerate(stream, start=1):
        yield element

        if i % step_size == 0:
            log.info("Processing status", processed=i, last=element)

        targets.add(element.target)
        if element.sink_path:
            sinks.add(element.sink_path)

    log.info("Pipeline stats", elements=i, targets=len(targets), sinks=len(sinks))


def sink_records(
    record_stream: Iterable[RecordStreamElement],
    state: DumpState,
) -> Iterator[RecordStreamElement]:
    """
    Persist records from the stream into appropriate sinks, per serialization, compression and record type.
    """
    with cached_sink_writers(state) as write_element:
        for element in record_stream:
            write_element(element)
            yield element


def persist_processing_state(
    record_stream: Iterable[RecordStreamElement],
    state: DumpState,
) -> Iterator[RecordStreamElement]:
    """
    Keep track of the pipeline state in a persistent state object.
    """
    with persisted_state(state) as save_state_periodically:
        for element in record_stream:
            save_state_periodically()
            yield element


def execute_pipeline(
    targets: list[str],
    functions: str,
    output_dir: Path,
    serialization: Serialization,
    compression: Optional[Compression] = None,
    restart: Optional[bool] = False,
    limit: Optional[int] = None,
) -> None:
    """
    Run the record generation, processing and sinking pipeline.
    """

    compression = compression or Compression.NONE

    state = None if restart else load_state(output_dir=output_dir)

    if state:
        log.info("Resuming state", output=output_dir, state=state.path)

        if (
            state.serialization != serialization
            or state.compression != compression
            or state.functions != functions
            or state.target_paths != targets
        ):
            log.error(
                "Configuration of the existing state conflicts with parameters provided",
                state_serialization=state.serialization,
                state_compression=state.compression,
                state_targets=state.target_paths,
                state_functions=state.functions,
                targets=targets,
                functions=functions,
                serialization=serialization,
                compression=compression,
                state_path=state.path,
            )
            return
    else:
        state = create_state(
            output_dir=output_dir,
            target_paths=targets,
            functions=functions,
            serialization=serialization,
            compression=compression,
        )
        log.info("New state created", restart=restart, state=state.path)

    targets_stream = get_targets(targets)
    target_func_pairs_stream = produce_target_func_pairs(targets_stream, functions, state)
    record_stream = execute_functions(target_func_pairs_stream)

    if limit:
        record_stream = itertools.islice(record_stream, limit)

    record_stream = sink_records(record_stream, state)
    record_stream = log_progress(record_stream)
    record_stream = persist_processing_state(record_stream, state)

    # exhaust the generator, executing all pipeline steps
    deque(record_stream, maxlen=0)

    log.info("Pipeline has finished")


def parse_arguments() -> argparse.Namespace:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        description="dissect.target",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
        add_help=True,
    )
    parser.add_argument("targets", metavar="TARGET", nargs="+", help="targets to load")
    parser.add_argument("-f", "--function", required=True, help="one or more comma separated functions to execute")

    parser.add_argument(
        "-c",
        "--compression",
        choices=[c.value for c in Compression if c is not Compression.NONE],
        help="compression method",
    )
    parser.add_argument(
        "--restart",
        action="store_true",
        help="restart the session and overwrite the state file if it exists",
    )

    parser.add_argument(
        "-s",
        "--serialization",
        choices=[s.value for s in Serialization],
        default=Serialization.JSONLINES.value,
        help="serialization method",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=".",
        required=True,
        help="output directory",
    )
    parser.add_argument("--limit", type=int, help="limit number of records produced")
    configure_generic_arguments(parser)

    args, _ = parser.parse_known_args()
    process_generic_arguments(args)

    return args


def main():
    args = parse_arguments()

    try:
        execute_pipeline(
            targets=args.targets,
            functions=args.function,
            output_dir=args.output,
            serialization=Serialization(args.serialization),
            compression=Compression(args.compression),
            restart=args.restart,
            limit=args.limit,
        )
    except Exception:
        log.error("Exception while running the pipeline", exc_info=True)
        sys.exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import zipfile
from pathlib import Path
from typing import TYPE_CHECKING

from dissect.target.loaders.dir import DirLoader, find_dirs, map_dirs
from dissect.target.plugin import OperatingSystem

if TYPE_CHECKING:
    from dissect.target import Target

log = logging.getLogger(__name__)

FILESYSTEMS_ROOT = "uploads"
UNIX_ACCESSORS = ["file", "auto"]
WINDOWS_ACCESSORS = ["mft", "ntfs", "lazy_ntfs", "ntfs_vss", "auto"]


def find_fs_directories(path: Path) -> tuple[OperatingSystem | None, list[Path] | None]:
    fs_root = path.joinpath(FILESYSTEMS_ROOT)

    # Unix
    for accessor in UNIX_ACCESSORS:
        accessor_root = fs_root.joinpath(accessor)
        if accessor_root.exists():
            os_type, dirs = find_dirs(accessor_root)
            if os_type in [OperatingSystem.UNIX, OperatingSystem.LINUX, OperatingSystem.OSX]:
                return os_type, [dirs[0]]

    # Windows
    volumes = set()
    vss_volumes = set()
    for accessor in WINDOWS_ACCESSORS:
        accessor_root = fs_root.joinpath(accessor)
        if accessor_root.exists():
            # If the accessor directory exists, assume all the subdirectories are volumes
            for volume in accessor_root.iterdir():
                if not volume.is_dir():
                    continue

                # https://github.com/Velocidex/velociraptor/blob/87368e7cc678144592a1614bb3bbd0a0f900ded9/accessors/ntfs/vss.go#L82
                if "HarddiskVolumeShadowCopy" in volume.name:
                    vss_volumes.add(volume)
                elif (drive_letter := extract_drive_letter(volume.name)) is not None:
                    volumes.add((drive_letter, volume))
                else:
                    volumes.add(volume)

    if volumes:
        # The volumes that represent drives (C, D) are mounted first,
        # otherwise one of the volume shadow copies could be detected as the root filesystem which results in errors.
        return OperatingSystem.WINDOWS, list(volumes) + list(vss_volumes)

    return None, None


def extract_drive_letter(name: str) -> str | None:
    # \\.\X: in URL encoding
    if len(name) == 14 and name.startswith("%5C%5C.%5C") and name.endswith("%3A"):
        return name[10].lower()

    # X: in URL encoding
    if len(name) == 4 and name.endswith("%3A"):
        return name[0].lower()


class VelociraptorLoader(DirLoader):
    """Load Rapid7 Velociraptor forensic image files.

    As of Velociraptor version 0.7.0 the structure of the Velociraptor Offline Collector varies by operating system.
    Generic.Collectors.File (Unix) uses the accessors file and auto. The loader supports the following configuration::

        {"Generic.Collectors.File":{"Root":"/","collectionSpec":"Glob\\netc/**\\nvar/log/**"}}

    Generic.Collectors.File (Windows) and Windows.KapeFiles.Targets (Windows) uses the accessors mft, ntfs, lazy_ntfs,
    ntfs_vss and auto. The loader supports a collection where multiple accessors were used.

    References:
        - https://www.rapid7.com/products/velociraptor/
        - https://docs.velociraptor.app/
        - https://github.com/Velocidex/velociraptor
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path)

        if path.suffix == ".zip":
            log.warning(
                f"Velociraptor target {path!r} is compressed, which will slightly affect performance. "
                "Consider uncompressing the archive and passing the uncompressed folder to Dissect."
            )
            self.root = zipfile.Path(path.open("rb"))
        else:
            self.root = path

    @staticmethod
    def detect(path: Path) -> bool:
        # The 'uploads' folder contains the data acquired
        # The 'results' folder contains information about the used Velociraptor artifacts e.g. Generic.Collectors.File
        # The 'uploads.json' file contains information about the collected files
        # Collection-HOSTNAME-TIMESTAMP/
        #   uploads/
        #   results/
        #   uploads.json
        #   [...] other files related to the collection
        if path.exists() and path.suffix == ".zip":  # novermin
            path = zipfile.Path(path.open("rb"))

        if path.joinpath(FILESYSTEMS_ROOT).exists() and path.joinpath("uploads.json").exists():
            _, dirs = find_fs_directories(path)
            return bool(dirs)

        return False

    def map(self, target: Target) -> None:
        os_type, dirs = find_fs_directories(self.root)
        if os_type == OperatingSystem.WINDOWS:
            # Velociraptor doesn't have the correct filenames for the paths "$J" and "$Secure:$SDS"
            map_dirs(
                target,
                dirs,
                os_type,
                usnjrnl_path="$Extend/$UsnJrnl%3A$J",
                sds_path="$Secure%3A$SDS",
            )
        else:
            map_dirs(target, dirs, os_type)



from __future__ import annotations

import re
from datetime import datetime, timezone, tzinfo
from pathlib import Path
from typing import Iterator

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.helpers.utils import year_rollover_helper
from dissect.target.plugin import Plugin, alias, export

MessagesRecord = TargetRecordDescriptor(
    "linux/log/messages",
    [
        ("datetime", "ts"),
        ("string", "daemon"),
        ("varint", "pid"),
        ("string", "message"),
        ("path", "source"),
    ],
)

DEFAULT_TS_LOG_FORMAT = "%b %d %H:%M:%S"
RE_TS = re.compile(r"(\w+\s{1,2}\d+\s\d{2}:\d{2}:\d{2})")
RE_DAEMON = re.compile(r"^[^:]+:\d+:\d+[^\[\]:]+\s([^\[:]+)[\[|:]{1}")
RE_PID = re.compile(r"\w\[(\d+)\]")
RE_MSG = re.compile(r"[^:]+:\d+:\d+[^:]+:\s(.*)$")
RE_CLOUD_INIT_LINE = re.compile(
    r"^(?P<ts>\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3}) - (?P<daemon>.*)\[(?P<log_level>\w+)\]\: (?P<message>.*)$"
)


class MessagesPlugin(Plugin):
    """Unix messages log plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.log_files = set(self._find_log_files())

    def _find_log_files(self) -> Iterator[Path]:
        log_dirs = ["/var/log/", "/var/log/installer/"]
        file_globs = ["syslog*", "messages*", "cloud-init.log*"]
        for log_dir in log_dirs:
            for glob in file_globs:
                yield from self.target.fs.path(log_dir).glob(glob)

    def check_compatible(self) -> None:
        if not self.log_files:
            raise UnsupportedPluginError("No log files found")

    @alias("syslog")
    @export(record=MessagesRecord)
    def messages(self) -> Iterator[MessagesRecord]:
        """Return contents of /var/log/messages*, /var/log/syslog* and cloud-init logs.

        Due to year rollover detection, the contents of the files are returned in reverse.

        The messages log file holds information about a variety of events such as the system error messages, system
        startups and shutdowns, change in the network configuration, etc. Aims to store valuable, non-debug and
        non-critical messages. This log should be considered the "general system activity" log.

        References:
            - https://geek-university.com/linux/var-log-messages-file/
            - https://www.geeksforgeeks.org/file-timestamps-mtime-ctime-and-atime-in-linux/
            - https://cloudinit.readthedocs.io/en/latest/development/logging.html#logging-command-output
        """

        tzinfo = self.target.datetime.tzinfo

        for log_file in self.log_files:
            if "cloud-init" in log_file.name:
                yield from self._parse_cloud_init_log(log_file, tzinfo)
                continue

            for ts, line in year_rollover_helper(log_file, RE_TS, DEFAULT_TS_LOG_FORMAT, tzinfo):
                daemon = dict(enumerate(RE_DAEMON.findall(line))).get(0)
                pid = dict(enumerate(RE_PID.findall(line))).get(0)
                message = dict(enumerate(RE_MSG.findall(line))).get(0, line)

                yield MessagesRecord(
                    ts=ts,
                    daemon=daemon,
                    pid=pid,
                    message=message,
                    source=log_file,
                    _target=self.target,
                )

    def _parse_cloud_init_log(self, log_file: Path, tzinfo: tzinfo | None = timezone.utc) -> Iterator[MessagesRecord]:
        """Parse a cloud-init.log file.

        Lines are structured in the following format:
        ``YYYY-MM-DD HH:MM:SS,000 - dhcp.py[DEBUG]: Received dhcp lease on IFACE for IP/MASK``

        NOTE: ``cloud-init-output.log`` files are not supported as they do not contain structured logs.

        Args:
            ``log_file``: path to cloud-init.log file.

        Returns: ``MessagesRecord``
        """

        ts_fmt = "%Y-%m-%d %H:%M:%S,%f"

        with open_decompress(log_file, "rt") as fh:
            for line in fh:
                if not (line := line.strip()):
                    continue

                if not (match := RE_CLOUD_INIT_LINE.match(line)):
                    self.target.log.warning("Could not match cloud-init log line in file: %s", log_file)
                    self.target.log.debug("No match for line '%s'", line)
                    continue

                values = match.groupdict()

                # Actual format is ``YYYY-MM-DD HH:MM:SS,000`` (asctime with milliseconds) but python has no strptime
                # operator for 3 digit milliseconds, so we convert and pad to six digit microseconds.
                # https://github.com/canonical/cloud-init/blob/main/cloudinit/log/loggers.py#DEFAULT_LOG_FORMAT
                # https://docs.python.org/3/library/logging.html#asctime
                raw_ts, _, milliseconds = values["ts"].rpartition(",")
                raw_ts += "," + str((int(milliseconds) * 1000)).zfill(6)

                try:
                    ts = datetime.strptime(raw_ts, ts_fmt).replace(tzinfo=tzinfo)

                except ValueError as e:
                    self.target.log.warning("Timestamp '%s' does not match format '%s'", raw_ts, ts_fmt)
                    self.target.log.debug("", exc_info=e)
                    ts = datetime(1970, 1, 1, 0, 0, 0, 0)

                yield MessagesRecord(
                    ts=ts,
                    daemon=values["daemon"],
                    pid=None,
                    message=values["message"],
                    source=log_file,
                    _target=self.target,
                )



import argparse
import errno
import inspect
import json
import os
import sys
import textwrap
import urllib
from datetime import datetime
from functools import wraps
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Type, Union

from dissect.target import Target
from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers import docs, keychain
from dissect.target.helpers.docs import get_docstring
from dissect.target.loader import LOADERS_BY_SCHEME
from dissect.target.plugin import (
    OSPlugin,
    Plugin,
    PluginFunction,
    find_plugin_functions,
    get_external_module_paths,
    load_modules_from_paths,
)
from dissect.target.tools.logging import configure_logging


def configure_generic_arguments(args_parser: argparse.ArgumentParser) -> None:
    args_parser.add_argument("-K", "--keychain-file", type=Path, help="keychain file in CSV format")
    args_parser.add_argument("-Kv", "--keychain-value", help="passphrase, recovery key or key file path value")
    args_parser.add_argument("-v", "--verbose", action="count", default=0, help="increase output verbosity")
    args_parser.add_argument("--version", action="store_true", help="print version")
    args_parser.add_argument("-q", "--quiet", action="store_true", help="do not output logging information")
    args_parser.add_argument(
        "--plugin-path",
        action="store",
        nargs="+",
        type=Path,
        help="a file or directory containing plugins and extensions",
    )


def process_generic_arguments(args: argparse.Namespace) -> None:
    configure_logging(args.verbose, args.quiet, as_plain_text=True)

    if args.version:
        try:
            print("dissect.target version " + version("dissect.target"))
        except PackageNotFoundError:
            print("unable to determine version")
        sys.exit(0)

    if args.keychain_file:
        keychain.register_keychain_file(args.keychain_file)

    if args.keychain_value:
        keychain.register_wildcard_value(args.keychain_value)

    paths = get_external_module_paths(args.plugin_path or [])
    load_modules_from_paths(paths)


def generate_argparse_for_bound_method(
    method: Callable,
    usage_tmpl: Optional[str] = None,
) -> argparse.ArgumentParser:
    """Generate an `argparse.ArgumentParser` for a bound `Plugin` class method"""

    # allow functools.partial wrapped method
    while hasattr(method, "func"):
        method = method.func

    if not inspect.ismethod(method):
        raise ValueError(f"Value `{method}` is not a bound plugin instance method")

    unbound_method = method.__func__
    return generate_argparse_for_unbound_method(unbound_method, usage_tmpl=usage_tmpl)


def generate_argparse_for_unbound_method(
    method: Callable,
    usage_tmpl: Optional[str] = None,
) -> argparse.ArgumentParser:
    """Generate an `argparse.ArgumentParser` for an unbound `Plugin` class method"""

    if not inspect.isfunction(method):
        raise ValueError(f"Value `{method}` is not an unbound plugin method")

    desc = method.__doc__ or docs.get_func_description(method, with_docstrings=True)

    if "\n" in desc:
        desc = inspect.cleandoc(desc)

    help_formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(description=desc, formatter_class=help_formatter, conflict_handler="resolve")

    fargs = getattr(method, "__args__", [])
    groups = {}
    default_group_options = {"required": False}
    for args, kwargs in fargs:
        if "group" in kwargs:
            group_name = kwargs.pop("group")
            options = kwargs.pop("group_options") if "group_options" in kwargs else default_group_options
            if group_name not in groups:
                group = parser.add_mutually_exclusive_group(**options)
                groups[group_name] = group
            else:
                group = groups[group_name]

            group.add_argument(*args, **kwargs)
        else:
            parser.add_argument(*args, **kwargs)

    usage = parser.format_usage()
    offset = usage.find(parser.prog) + len(parser.prog)
    func_name = method.__name__
    usage_tmpl = usage_tmpl or "{prog} {usage}"
    parser.usage = usage_tmpl.format(prog=parser.prog, name=func_name, usage=usage[offset:])

    return parser


def generate_argparse_for_plugin_class(
    plugin_cls: Type[Plugin],
    usage_tmpl: Optional[str] = None,
) -> argparse.ArgumentParser:
    """Generate an `argparse.ArgumentParser` for a `Plugin` class"""

    if not isinstance(plugin_cls, type) or not issubclass(plugin_cls, Plugin):
        raise ValueError(f"`plugin_cls` must be a valid plugin class, not `{plugin_cls}`")

    method_name = plugin_cls.__namespace__
    desc = docs.get_plugin_overview(plugin_cls, with_plugin_desc=True, with_func_docstrings=True)

    help_formatter = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(description=desc, formatter_class=help_formatter, conflict_handler="resolve")

    usage = parser.format_usage()
    offset = usage.find(parser.prog) + len(parser.prog)
    usage_tmpl = usage_tmpl or "{prog} {usage}"

    parser.usage = usage_tmpl.format(prog=parser.prog, name=method_name, usage=usage[offset:])

    return parser


def generate_argparse_for_plugin(
    plugin_instance: Plugin,
    usage_tmpl: Optional[str] = None,
) -> argparse.ArgumentParser:
    """Generate an `argparse.ArgumentParser` for a `Plugin` instance"""

    if not isinstance(plugin_instance, Plugin):
        raise ValueError(f"`plugin_instance` must be a valid plugin instance, not `{plugin_instance}`")

    if not hasattr(plugin_instance, "__namespace__"):
        raise ValueError(f"Plugin `{plugin_instance}` is not a namespace plugin and is not callable")

    return generate_argparse_for_plugin_class(plugin_instance.__class__, usage_tmpl=usage_tmpl)


def plugin_factory(
    target: Target, plugin: Union[type, object], funcname: str, namespace: Optional[str]
) -> tuple[Plugin, str]:
    if hasattr(target._loader, "instance"):
        return target.get_function(funcname, namespace=namespace)

    if isinstance(plugin, type):
        plugin_obj = plugin(target)
        target_attr = getattr(plugin_obj, funcname)
        return plugin_obj, target_attr
    else:
        return plugin, getattr(plugin, funcname)


def execute_function_on_target(
    target: Target,
    func: PluginFunction,
    cli_params: Optional[List[str]] = None,
) -> Tuple[str, Any, List[str]]:
    """
    Execute function `func` on provided target `target` with provided `cli_params` list.
    """

    cli_params = cli_params or []

    target_attr = get_target_attribute(target, func)
    plugin_method, parser = plugin_function_with_argparser(target_attr)

    if parser:
        parsed_params, cli_params = parser.parse_known_args(cli_params)
        method_kwargs = vars(parsed_params)
        value = plugin_method(**method_kwargs)
    else:
        value = target_attr

    output_type = getattr(plugin_method, "__output__", "default") if plugin_method else "default"
    return (output_type, value, cli_params)


def get_target_attribute(target: Target, func: PluginFunction) -> Union[Plugin, Callable]:
    """Retrieves the function attribute from the target.

    If the function does not exist yet, it will attempt to load it into the target.

    Args:
        target: The target we wish to run the function on.
        func: The function to run on the target.

    Returns:
        The function, either plugin or a callable to execute.

    Raises:
        UnsupportedPluginError: When the function was incompatible with the target.
    """
    plugin_class = func.class_object
    if ns := getattr(func, "plugin_desc", {}).get("namespace", None):
        plugin_class = getattr(target, ns)
    elif target.has_function(func.method_name):
        # If the function is already attached, use the one inside the target.
        plugin_class, _ = target.get_function(func.method_name)
    elif issubclass(plugin_class, OSPlugin):
        # OS plugin does not need to be added
        plugin_class = target._os_plugin
    else:
        try:
            target.add_plugin(plugin_class)
        except UnsupportedPluginError as e:
            raise UnsupportedPluginError(
                f"Unsupported function `{func.method_name}` for target with plugin {func.class_object}"
            ) from e

    _, target_attr = plugin_factory(target, plugin_class, func.method_name, func.plugin_desc["namespace"])
    return target_attr


def plugin_function_with_argparser(
    target_attr: Union[Plugin, Callable],
) -> tuple[Optional[Iterator], Optional[argparse.ArgumentParser]]:
    """Resolves which plugin function to execute, and creates the argument parser for said plugin."""
    plugin_method = None
    parser = None

    # goes first because plugins are callable
    if isinstance(target_attr, Plugin):
        plugin_obj = target_attr

        if not plugin_obj.__namespace__:
            raise ValueError(f"Plugin {plugin_obj} is not callable")

        plugin_method = plugin_obj.get_all_records
        parser = generate_argparse_for_plugin(plugin_obj)
    elif callable(target_attr):
        plugin_method = target_attr
        parser = generate_argparse_for_bound_method(target_attr)
    return plugin_method, parser


def persist_execution_report(output_dir: Path, report_data: Dict, timestamp: datetime) -> Path:
    timestamp = timestamp.strftime("%Y-%m-%d-%H%M%S")
    report_filename = f"target-report-{timestamp}.json"
    report_full_path = output_dir / report_filename
    report_full_path.write_text(json.dumps(report_data, sort_keys=True, indent=4))
    return report_full_path


def catch_sigpipe(func: Callable) -> Callable:
    """Catches KeyboardInterrupt and BrokenPipeError (OSError 22 on Windows)."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            print("Aborted!", file=sys.stderr)
            return 1
        except OSError as e:
            # Only catch BrokenPipeError or OSError 22
            if e.errno in (errno.EPIPE, errno.EINVAL):
                devnull = os.open(os.devnull, os.O_WRONLY)
                os.dup2(devnull, sys.stdout.fileno())
                return 1
            # Raise other exceptions
            raise

    return wrapper


def args_to_uri(targets: list[str], loader_name: str, rest: list[str]) -> list[str]:
    """Converts argument-style ``-L`` to URI-style.

    Turns:
        ``target-query /evtxs/* -L log --log-hint="evtx" -f evtx``
    into:
        ``target-query "log:///evtxs/*?hint=evtx" -f evtx``

    For loaders providing ``@arg()`` arguments.
    """
    loader = LOADERS_BY_SCHEME.get(loader_name, None)

    parser = argparse.ArgumentParser(
        argument_default=argparse.SUPPRESS, description="\n".join(textwrap.wrap(get_docstring(loader)))
    )
    for load_arg in getattr(loader, "__args__", []):
        parser.add_argument(*load_arg[0], **load_arg[1])
    args = vars(parser.parse_known_args(rest)[0])
    uris = []
    for target in targets:
        uris.append(f"{loader_name}://{target}" + (("?" + urllib.parse.urlencode(args)) if args else ""))
    return uris


def find_and_filter_plugins(
    target: Target, functions: str, excluded_func_paths: set[str] = None
) -> Iterator[PluginFunction]:
    # Keep a set of plugins that were already executed on the target.
    executed_plugins = set()
    excluded_func_paths = excluded_func_paths or set()

    func_defs, _ = find_plugin_functions(target, functions, compatibility=False)

    for func_def in func_defs:
        if func_def.path in excluded_func_paths:
            continue

        # Avoid executing same plugin for multiple OSes (like hostname)
        if func_def.name in executed_plugins:
            continue

        executed_plugins.add(func_def.name)

        yield func_def



import pytest

from dissect.target import volume
from dissect.target.helpers import keychain

try:
    from dissect.target.volumes.bde import BitlockerVolumeSystem

    HAS_DISSECT_FVE = True
except ModuleNotFoundError:
    HAS_DISSECT_FVE = False


from tests._utils import absolute_path


@pytest.fixture
def encrypted_volume():
    data_file = "_data/volumes/bde/enc-volume.bin"
    with open(absolute_path(data_file), "rb") as f:
        yield f


@pytest.mark.skipif(not HAS_DISSECT_FVE, reason="requires dissect.fve")
def test_bde_volume_failure(target_win, encrypted_volume):
    enc_vol = volume.Volume(encrypted_volume, 1, 0, None, None, None, disk=encrypted_volume)
    target_win.volumes.add(enc_vol)
    target_win.volumes.apply()

    assert len(target_win.volumes) == 1
    assert enc_vol in target_win.volumes


@pytest.mark.skipif(not HAS_DISSECT_FVE, reason="requires dissect.fve")
def test_bde_volume_with_recovery_key(target_win, encrypted_volume):
    recovery_key = "272316-265804-640728-713570-509047-503305-045837-324731"

    keychain.register_key(
        keychain.KeyType.RECOVERY_KEY,
        recovery_key,
        identifier=None,
        provider=BitlockerVolumeSystem.__type__,
    )

    enc_vol = volume.Volume(encrypted_volume, 1, 0, None, None, None, disk=encrypted_volume)
    target_win.volumes.add(enc_vol)
    target_win.volumes.apply()

    assert len(target_win.volumes) == 2
    assert enc_vol in target_win.volumes

    dec_vol = next(v for v in target_win.volumes if v != enc_vol)

    # virtual fs + ntfs fs
    assert len(target_win.filesystems) == 2
    target_win.fs.mount("e:/", dec_vol.fs)

    assert target_win.fs.path("e:/test-folder/test-file-2.txt").exists()


@pytest.mark.skipif(not HAS_DISSECT_FVE, reason="requires dissect.fve")
def test_bde_volume_with_passphrase(target_win, encrypted_volume):
    identifier = "B6AD258A-2725-4A42-93C6-844478BF7A90"
    passphrase = "Password1234"

    keychain.register_key(
        keychain.KeyType.PASSPHRASE,
        passphrase,
        identifier=identifier,
        provider=BitlockerVolumeSystem.__type__,
    )

    enc_vol = volume.Volume(encrypted_volume, 1, 0, None, None, None, disk=encrypted_volume)
    target_win.volumes.add(enc_vol)
    target_win.volumes.apply()

    assert len(target_win.volumes) == 2
    assert enc_vol in target_win.volumes

    dec_vol = next(v for v in target_win.volumes if v != enc_vol)

    # virtual fs + ntfs fs
    assert len(target_win.filesystems) == 2
    target_win.fs.mount("e:/", dec_vol.fs)

    assert target_win.fs.path("e:/test-folder/test-file-2.txt").exists()


@pytest.mark.skipif(not HAS_DISSECT_FVE, reason="requires dissect.fve")
def test_bde_volume_with_wildcard_key(target_win, encrypted_volume):
    keychain.register_wildcard_value("Password1234")

    enc_vol = volume.Volume(encrypted_volume, 1, 0, None, None, None, disk=encrypted_volume)
    target_win.volumes.add(enc_vol)
    target_win.volumes.apply()

    assert len(target_win.volumes) == 2
    assert enc_vol in target_win.volumes

    dec_vol = next(v for v in target_win.volumes if v != enc_vol)

    # virtual fs + ntfs fs
    assert len(target_win.filesystems) == 2
    target_win.fs.mount("e:/", dec_vol.fs)

    assert target_win.fs.path("e:/test-folder/test-file-2.txt").exists()



from typing import Iterator

from dissect.clfs import blf, container
from dissect.clfs.exceptions import InvalidBLFError, InvalidRecordBlockError

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers import fsutil
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export
from dissect.target.target import Target

ClfsRecord = TargetRecordDescriptor(
    "filesystem/windows/clfs",
    [
        ("string", "stream_name"),
        ("uint32", "stream_id"),
        ("string", "type"),
        ("string", "file_attributes"),
        ("uint32", "offset"),
        ("string", "container_name"),
        ("uint32", "container_id"),
        ("uint32", "container_size"),
        ("uint32", "record_offset"),
        ("bytes", "record_data"),
        ("bytes", "block_data"),
    ],
)


class ClfsPlugin(Plugin):
    """CLFS Plugin.

    Dissect plugin for parsing the Base Log Files of a Microsoft Windows system.

    Most of these records are actually parsed in-memory, this is the first iteration
    to parse the files present on disk. This should be improved in the near future when
    the memory implementation for dissect is working.
    """

    BLF_PATH = "sysvol/windows/system32/config/"  # Unsure at time of writing if this is the only location

    def __init__(self, target: Target):
        super().__init__(target)
        self._blfs = []

        blfdir = self.target.fs.path(self.BLF_PATH)

        if blfdir.exists() and blfdir.is_dir():
            blf_files = blfdir.glob("*.blf")

            for blf_path in blf_files:
                fh = blf_path.open()

                try:
                    blf_instance = blf.BLF(fh)
                    self._blfs.append((blf_path, blf_instance))
                except InvalidRecordBlockError as e:
                    self.target.log.warning(f"Invalid record block: {blf_path}", exc_info=e)
                except InvalidBLFError as e:
                    self.target.log.warning(f"Could not validate BLF: {blf_path}", exc_info=e)

    def check_compatible(self) -> None:
        if not self._blfs:
            raise UnsupportedPluginError("No BLF files found")

    @export(record=ClfsRecord)
    def clfs(self) -> Iterator[ClfsRecord]:
        """Parse the containers associated with a valid BLF file.

        Containers are used to store the transactional logs in the form of records.

        References:
            - https://docs.microsoft.com/en-us/windows-hardware/drivers/kernel/introduction-to-the-common-log-file-system
        """  # noqa: E501

        for blf_path, blf_instance in self._blfs:
            # We only parse the base record client/container contexts for now
            for base_record in blf_instance.base_records():
                for stream in base_record.streams:
                    for blf_container in base_record.containers:
                        # Check if the stream ID is matching the container ID
                        if blf_container.id != stream.lsn_base.Offset.ContainerId:
                            continue

                        # We can encounter the same container ID for the shadow blocks
                        if blf_container.type != stream.type:
                            continue

                        # Invalid LSN (-1)
                        if stream.lsn_base.PhysicalOffset <= 0:
                            continue

                        container_path = blf_container.name.replace("%BLF%", str(blf_path.parent))
                        container_path = fsutil.normalize(container_path, alt_separator=blf_path._flavour.altsep)
                        container_file = self.target.fs.path(container_path)

                        fh = container_file.open()
                        trans = container.Container(fh, offset=stream.offset)

                        # Open each container and yield the results for each record found within that container
                        for record_offset, record_data, block_data in trans.records():
                            yield ClfsRecord(
                                stream_name=stream.name,
                                stream_id=stream.id,
                                type=stream.type,
                                file_attributes=stream.file_attributes,
                                offset=stream.offset,
                                container_name=container_file.name,
                                container_id=stream.lsn_base.Offset.ContainerId,
                                container_size=blf_container.size,
                                record_offset=record_offset,
                                record_data=record_data,
                                block_data=block_data,
                                _target=self.target,
                            )



from io import BytesIO

import pytest

from dissect.target import container
from dissect.target.containers.split import SplitContainer


def _assert_split_container(fh: SplitContainer):
    assert isinstance(fh, SplitContainer)
    assert fh.read(4096) == (b"A" * 512) + (b"B" * 512) + (b"C" * 512) + (b"D" * 512)


@pytest.fixture
def split_fhs():
    fhs = []
    for char in b"ABCD":
        fhs.append(BytesIO(bytes([char] * 512)))

    yield fhs


@pytest.fixture
def split_paths(tmp_path, split_fhs):
    paths = [(tmp_path / f"split.{i:>03}") for i in range(4)]

    for fh, path in zip(split_fhs, paths):
        fh.seek(0)
        path.write_bytes(fh.read())

    yield paths


@pytest.fixture
def split_symlink(tmp_path, split_paths):
    dir_path = tmp_path / "dir"
    symlink_path = dir_path / split_paths[0].name
    dir_path.mkdir()
    symlink_path.symlink_to(split_paths[0])

    yield symlink_path


@pytest.mark.parametrize(
    "obj",
    [
        "split_fhs",
        "split_paths",
        "split_symlink",
    ],
)
def test_split_container(obj, request):
    fh = container.open(request.getfixturevalue(obj))
    _assert_split_container(fh)



from __future__ import annotations

from typing import List, Optional

from dissect.target.filesystem import Filesystem
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix._os import UnixPlugin
from dissect.target.target import Target


class BsdPlugin(UnixPlugin):
    def __init__(self, target: Target):
        super().__init__(target)

    @classmethod
    def detect(cls, target: Target) -> Optional[Filesystem]:
        for fs in target.filesystems:
            # checking the existence of /var/authpf for free- and openbsd
            # checking the existence of /var/at for net- and freebsd
            if fs.exists("/usr/obj") or fs.exists("/altroot") or fs.exists("/etc/pf.conf"):
                return fs

        return None

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.BSD.value

    @export(property=True)
    def hostname(self) -> Optional[str]:
        fh = self.target.fs.path("/etc/rc.conf")

        for line in fh.open("rt").readlines():
            if line.startswith("hostname"):
                hostname = line.rstrip().split("=", maxsplit=1)[1].replace('"', "")
                return hostname

    @export(property=True)
    def ips(self) -> Optional[List[str]]:
        self.target.log.error(f"ips plugin not implemented for {self.__class__}")
        return None



"""Dissect plugin system.

See ``dissect/target/plugins/general/example.py`` for an example plugin.
"""

from __future__ import annotations

import fnmatch
import functools
import importlib
import importlib.util
import inspect
import logging
import os
import sys
import traceback
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Iterator, Optional, Type

from flow.record import Record, RecordDescriptor

import dissect.target.plugins.general as general
from dissect.target.exceptions import PluginError, UnsupportedPluginError
from dissect.target.helpers import cache
from dissect.target.helpers.record import EmptyRecord
from dissect.target.helpers.utils import StrEnum

try:
    from dissect.target.plugins._pluginlist import PLUGINS

    GENERATED = True
except Exception:
    PLUGINS = {}
    GENERATED = False

if TYPE_CHECKING:
    from dissect.target import Target
    from dissect.target.filesystem import Filesystem
    from dissect.target.helpers.record import ChildTargetRecord

PluginDescriptor = dict[str, Any]
"""A dictionary type, for what the plugin descriptor looks like."""

MODULE_PATH = "dissect.target.plugins"
"""The base module path to the in-tree plugins."""
OUTPUTS = (
    "default",
    "record",
    "yield",
    "none",
)
"""The different output types supported by ``@export``."""

log = logging.getLogger(__name__)


class OperatingSystem(StrEnum):
    ANDROID = "android"
    BSD = "bsd"
    CITRIX = "citrix-netscaler"
    ESXI = "esxi"
    FORTIOS = "fortios"
    IOS = "ios"
    LINUX = "linux"
    OSX = "osx"
    PROXMOX = "proxmox"
    UNIX = "unix"
    VYOS = "vyos"
    WINDOWS = "windows"


def export(*args, **kwargs) -> Callable:
    """Decorator to be used on Plugin functions that should be exported.

    Supported keyword arguments:
        property (bool): Whether this export should be regarded as a property.
            Properties are implicitly cached.

        cache (bool): Whether the result of this function should be cached.

        record (RecordDescriptor): The :class:`flow.record.RecordDescriptor` for the records that this function yields.
            If the records are dynamically made, use DynamicRecord instead.

        output (str): The output type of this function. Must be one of:
        - default: Single return value
        - record: Yields records. Implicit when record argument is given.
        - yield: Yields printable values.
        - none: No return value. Plugin is responsible for output formatting and should return ``None``.

    The ``export`` decorator adds some additional private attributes to an exported method or property:
    - ``__output__``: The output type to expect for this function, this is the same as ``output``.
    - ``__record__``: The type of record to expect, this value is the same as ``record``.
    - ``__exported__``: set to ``True`` to indicate the method or property is exported.

    Raises:
        ValueError: if there was an invalid output type.

    Returns:
        An exported function from a plugin.
    """

    def decorator(obj):
        # Properties are implicitly cached
        # Important! Currently it's crucial that this is *always* called
        # See the comment in Plugin.__init_subclass__ for more detail regarding Plugin.get_all_records
        obj = cache.wrap(obj, no_cache=not kwargs.get("cache", True), cls=kwargs.get("cls", None))

        output = kwargs.get("output", "default")
        if output not in OUTPUTS:
            options = ", ".join(OUTPUTS)
            raise ValueError(f'Invalid output method "{output}", must be one of {options}')

        record = kwargs.get("record", None)
        if record is not None:
            output = "record"

        obj.__output__ = output
        obj.__record__ = record
        obj.__exported__ = True

        if kwargs.get("property", False):
            obj = property(obj)

        return obj

    if len(args) == 1:
        return decorator(args[0])
    else:
        return decorator


def get_nonprivate_attribute_names(cls: Type[Plugin]) -> list[str]:
    """Retrieve all attributes that do not start with ``_``."""
    return [attr for attr in dir(cls) if not attr.startswith("_")]


def get_nonprivate_attributes(cls: Type[Plugin]) -> list[Any]:
    """Retrieve all public attributes of a :class:`Plugin`."""
    # Note: `dir()` might return attributes from parent class
    return [getattr(cls, attr) for attr in get_nonprivate_attribute_names(cls)]


def get_nonprivate_methods(cls: Type[Plugin]) -> list[Callable]:
    """Retrieve all public methods of a :class:`Plugin`."""
    return [attr for attr in get_nonprivate_attributes(cls) if not isinstance(attr, property) and callable(attr)]


def get_descriptors_on_nonprivate_methods(cls: Type[Plugin]) -> list[RecordDescriptor]:
    """Return record descriptors set on nonprivate methods in `cls` class."""
    descriptors = set()
    methods = get_nonprivate_methods(cls)

    for m in methods:
        if not hasattr(m, "__record__"):
            continue

        record = m.__record__
        if not record:
            continue

        try:
            # check if __record__ value is iterable (for example, a list)
            descriptors.update(record)
        except TypeError:
            descriptors.add(record)
    return list(descriptors)


class Plugin:
    """Base class for plugins.

    Plugins can optionally be namespaced by specifying the ``__namespace__``
    class attribute. Namespacing results in your plugin needing to be prefixed
    with this namespace when being called. For example, if your plugin has
    specified ``test`` as namespace and a function called ``example``, you must
    call your plugin with ``test.example``.

    A ``Plugin`` class has the following private class attributes:

    - ``__namespace__``
    - ``__record_descriptors__``

    With the following three being assigned in :func:`register`:

    - ``__plugin__``
    - ``__functions__``
    - ``__exports__``

    Additionally, the methods and attributes of :class:`Plugin` receive more private attributes
    by using decorators.

    The :func:`export` decorator adds the following private attributes

    - ``__exported__``
    - ``__output__``: Set with the :func:`export` decorator.
    - ``__record__``: Set with the :func:`export` decorator.

    The :func:`internal` decorator and :class:`InternalPlugin` set the ``__internal__`` attribute.
    Finally. :func:`args` decorator sets the ``__args__`` attribute.

    The :func:`alias` decorator populates the ``__aliases__`` private attribute of :class:`Plugin` methods.

    Args:
        target: The :class:`~dissect.target.target.Target` object to load the plugin for.
    """

    __namespace__: str = None
    """Defines the plugin namespace."""
    __record_descriptors__: list[RecordDescriptor] = None
    """Defines a list of :class:`~flow.record.RecordDescriptor` of the exported plugin functions."""
    __register__: bool = True
    """Determines whether this plugin will be registered."""
    __findable__: bool = True
    """Determines whether this plugin will be revealed when using search patterns.

    Some (meta)-plugins are not very suitable for wild cards on CLI or
    plugin searches, because they will produce duplicate records or results.
    For instance a plugin that offers the same functions as subplugins will
    produce redundant results when used with a wild card
    (browser.* -> browser.history + browser.*.history).
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        # Do not register the "base" subclassess `OSPlugin` and `ChildTargetPlugin`
        if cls.__name__ not in ("OSPlugin", "ChildTargetPlugin") and cls.__register__:
            register(cls)

        record_descriptors = get_descriptors_on_nonprivate_methods(cls)
        cls.__record_descriptors__ = record_descriptors
        # This is a bit tricky currently
        # cls.get_all_records is the *function* Plugin.get_all_records, not from the subclass
        # export() currently will _always_ return a new object because it always calls ``cache.wrap(obj)``
        # This allows this to work, otherwise the Plugin.get_all_records would get all the plugin attributes set on it
        cls.get_all_records = export(output="record", record=record_descriptors, cache=False, cls=cls)(
            cls.get_all_records
        )

    def __init__(self, target: Target):
        self.target = target

    def is_compatible(self) -> bool:
        """Perform a compatibility check with the target."""
        try:
            self.check_compatible()
        except Exception:
            return False

        return True

    def check_compatible(self) -> None:
        """Perform a compatibility check with the target.

        This function should return ``None`` if the plugin is compatible with
        the current target (``self.target``). For example, check if a certain
        file exists.
        Otherwise it should raise an ``UnsupportedPluginError``.

        Raises:
            UnsupportedPluginError: If the plugin could not be loaded.
        """
        raise NotImplementedError

    def get_all_records(self) -> Iterator[Record]:
        """Return the records of all exported methods.

        Raises:
            PluginError: If the subclass is not a namespace plugin.
        """
        if not self.__namespace__:
            raise PluginError(f"Plugin {self.__class__.__name__} is not a namespace plugin")

        for method_name in self.__exports__:
            method = getattr(self, method_name)

            try:
                yield from method()
            except Exception:
                self.target.log.error("Error while executing `%s.%s`", self.__namespace__, method_name, exc_info=True)

    def __call__(self, *args, **kwargs):
        """A shortcut to :func:`get_all_records`.

        Raises:
            PluginError: If the subclass is not a namespace plugin.
        """
        if not self.__namespace__:
            raise PluginError(f"Plugin {self.__class__.__name__} is not a callable")
        return self.get_all_records()


class OSPlugin(Plugin):
    """Base class for OS plugins.

    This provides a base class for certain common functions of OS's, which each OS plugin has to implement separately.

    For example, it provides an interface for retrieving the hostname and users of a target.

    All derived classes MUST implement ALL the classmethods and exported
    methods with the same ``@classmethod`` or ``@export(...)`` annotation.
    """

    def __init_subclass__(cls, **kwargs):
        # Note that cls is the subclass
        super().__init_subclass__(**kwargs)

        for os_method in get_nonprivate_attributes(OSPlugin):
            if isinstance(os_method, property):
                os_method = os_method.fget
            os_docstring = os_method.__doc__

            method = getattr(cls, os_method.__name__, None)
            if isinstance(method, property):
                method = method.fget
            # This works as None has a __doc__ property (which is None).
            docstring = method.__doc__

            if method and not docstring:
                if hasattr(method, "__func__"):
                    method = method.__func__
                method.__doc__ = os_docstring

    def check_compatible(self) -> bool:
        """OSPlugin's use a different compatibility check, override the one from the :class:`Plugin` class.

        Returns:
            This function always returns ``True``.
        """
        return True

    @classmethod
    def detect(cls, fs: Filesystem) -> Optional[Filesystem]:
        """Provide detection of this OSPlugin on a given filesystem.

        Args:
            fs: :class:`~dissect.target.filesystem.Filesystem` to detect the OS on.

        Returns:
            The root filesystem / sysvol when found.
        """
        raise NotImplementedError

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> OSPlugin:
        """Initiate this OSPlugin with the given target and detected filesystem.

        Args:
            target: The :class:`~dissect.target.target.Target` object.
            sysvol: The filesystem that was detected in the ``detect()`` function.

        Returns:
            An instantiated version of the OSPlugin.
        """
        raise NotImplementedError

    @export(property=True)
    def hostname(self) -> Optional[str]:
        """Return the target's hostname.

        Returns:
            The hostname as string.
        """
        raise NotImplementedError

    @export(property=True)
    def ips(self) -> list[str]:
        """Return the IP addresses configured in the target.

        Returns:
            The IPs as list.
        """
        raise NotImplementedError

    @export(property=True)
    def version(self) -> Optional[str]:
        """Return the target's OS version.

        Returns:
            The OS version as string.
        """
        raise NotImplementedError

    @export(record=EmptyRecord)
    def users(self) -> list[Record]:
        """Return the users available in the target.

        Returns:
            A list of user records.
        """
        raise NotImplementedError

    @export(property=True)
    def os(self) -> str:
        """Return a slug of the target's OS name.

        Returns:
            A slug of the OS name, e.g. 'windows' or 'linux'.
        """
        raise NotImplementedError

    @export(property=True)
    def architecture(self) -> Optional[str]:
        """Return a slug of the target's OS architecture.

        Returns:
            A slug of the OS architecture, e.g. 'x86_32-unix', 'MIPS-linux' or
            'AMD64-win32', or 'unknown' if the architecture is unknown.
        """
        raise NotImplementedError


class ChildTargetPlugin(Plugin):
    """A Child target is a special plugin that can list more Targets.

    For example, :class:`~dissect.target.plugins.child.esxi.ESXiChildTargetPlugin` can
    list all of the Virtual Machines on the host.
    """

    __type__ = None

    def list_children(self) -> Iterator[ChildTargetRecord]:
        """Yield :class:`~dissect.target.helpers.record.ChildTargetRecord` records of all
        possible child targets on this target.
        """
        raise NotImplementedError


def register(plugincls: Type[Plugin]) -> None:
    """Register a plugin, and put related data inside :attr:`PLUGINS`.

    This function uses the following private attributes that are set using decorators:
        - ``__exported__``: Set in :func:`export`.
        - ``__internal__``: Set in :func:`internal`.

    Additionally, ``register`` sets the following private attributes on the `plugincls`:
        - ``__plugin__``: Always set to ``True``.
        - ``__functions__``: A list of all the methods and properties that are ``__internal__`` or ``__exported__``.
        - ``__exports__``: A list of all the methods or properties that were explicitly exported.

    Args:
        plugincls: A plugin class to register.

    Raises:
        ValueError: If ``plugincls`` is not a subclass of :class:`Plugin`.
    """
    if not issubclass(plugincls, Plugin):
        raise ValueError("Not a subclass of Plugin")

    exports = []
    functions = []

    # First pass to resolve aliases
    for attr in get_nonprivate_attributes(plugincls):
        for alias in getattr(attr, "__aliases__", []):
            clone_alias(plugincls, attr, alias)

    for attr in get_nonprivate_attributes(plugincls):
        if isinstance(attr, property):
            attr = attr.fget

        if getattr(attr, "__autogen__", False) and plugincls != plugincls.__nsplugin__:
            continue

        if getattr(attr, "__exported__", False):
            exports.append(attr.__name__)
            functions.append(attr.__name__)

        if getattr(attr, "__internal__", False):
            functions.append(attr.__name__)

    plugincls.__plugin__ = True
    plugincls.__functions__ = functions
    plugincls.__exports__ = exports

    modpath = _modulepath(plugincls)
    lookup_path = modpath
    if modpath.endswith("._os"):
        lookup_path, _, _ = modpath.rpartition(".")

    root = _traverse(lookup_path, PLUGINS)

    log.debug("Plugin registered: %s.%s", plugincls.__module__, plugincls.__qualname__)

    if issubclass(plugincls, (OSPlugin, ChildTargetPlugin)):
        if issubclass(plugincls, OSPlugin):
            special_key = "_os"
        elif issubclass(plugincls, ChildTargetPlugin):
            special_key = "_child"

        root[special_key] = {}
        root = root[special_key]

    # Check if the plugin was already registered
    if "class" in root and root["class"] == plugincls.__name__:
        return

    # Finally register the plugin
    root["class"] = plugincls.__name__
    root["module"] = modpath
    root["functions"] = plugincls.__functions__
    root["exports"] = plugincls.__exports__
    root["namespace"] = plugincls.__namespace__
    root["fullname"] = ".".join((plugincls.__module__, plugincls.__qualname__))
    root["is_osplugin"] = issubclass(plugincls, OSPlugin)


def internal(*args, **kwargs) -> Callable:
    """Decorator to be used on plugin functions that should be internal only.

    Making a plugin internal means that it's only callable from the Python API and not through ``target-query``.

    This decorator adds the ``__internal__`` private attribute to a method or property.
    The attribute is always set to ``True``, to tell :func:`register` that it is an internal
    method or property.
    """

    def decorator(obj):
        obj.__internal__ = True
        if kwargs.get("property", False):
            obj = property(obj)
        return obj

    if len(args) == 1:
        return decorator(args[0])
    else:
        return decorator


def arg(*args, **kwargs) -> Callable:
    """Decorator to be used on Plugin functions that accept additional command line arguments.

    Command line arguments can be added using the ``@arg`` decorator.
    Arguments to this decorator are directly forwarded to the ``ArgumentParser.add_argument`` function of ``argparse``.
    Resulting arguments are passed to the function using kwargs.
    The keyword argument name must match the argparse argument name.

    This decorator adds the ``__args__`` private attribute to a method or property.
    This attribute holds all the command line arguments that were added to the plugin function.
    """

    def decorator(obj):
        if not hasattr(obj, "__args__"):
            obj.__args__ = []
        arglist = getattr(obj, "__args__", [])
        arglist.append((args, kwargs))
        return obj

    return decorator


def alias(*args, **kwargs: dict[str, Any]) -> Callable:
    """Decorator to be used on :class:`Plugin` functions to register an alias of that function."""

    if not kwargs.get("name") and not args:
        raise ValueError("Missing argument 'name'")

    def decorator(obj: Callable) -> Callable:
        if not hasattr(obj, "__aliases__"):
            obj.__aliases__ = []

        if name := (kwargs.get("name") or args[0]):
            obj.__aliases__.append(name)

        return obj

    return decorator


def clone_alias(cls: type, attr: Callable, alias: str) -> None:
    """Clone the given attribute to an alias in the provided class."""

    # Clone the function object
    clone = type(attr)(attr.__code__, attr.__globals__, alias, attr.__defaults__, attr.__closure__)
    clone.__kwdefaults__ = attr.__kwdefaults__

    # Copy some attributes
    functools.update_wrapper(clone, attr)
    if wrapped := getattr(attr, "__wrapped__", None):
        # update_wrapper sets a new wrapper, we want the original
        clone.__wrapped__ = wrapped

    # Update module path so we can fool inspect.getmodule with subclassed Plugin classes
    clone.__module__ = cls.__module__

    # Update the names
    clone.__name__ = alias
    clone.__qualname__ = f"{cls.__name__}.{alias}"

    setattr(cls, alias, clone)


def plugins(
    osfilter: Optional[type[OSPlugin]] = None,
    special_keys: set[str] = set(),
    only_special_keys: bool = False,
) -> Iterator[PluginDescriptor]:
    """Walk the ``PLUGINS`` tree and return plugins.

    If ``osfilter`` is specified, only plugins related to the provided
    OSPlugin, or plugins with no OS relation are returned.
    If ``osfilter`` is ``None``, all plugins will be returned.

    One exception to this is if the ``osfilter`` is a (sub-)class of
    DefaultPlugin, then plugins are returned as if no ``osfilter`` was
    specified.

    Another exeption to this are plugins in the ``PLUGINS`` tree which are
    under a key that starts with a '_'. Those are only returned if their exact
    key is specified in ``special_keys``.

    An exception to these exceptions is in the case of ``OSPlugin`` (sub-)class
    plugins and ``os_filter`` is not ``None``. These plugins live in the
    ``PLUGINS`` tree under the ``_os`` special key. Those plugins are only
    returned if they fully match the provided ``osfilter``.

    The ``only_special_keys`` option returns only the plugins which are under a
    special key that is defined in ``special_keys``. All filtering here will
    happen as stated in the above cases.

    Args:
        osfilter: The optional OSPlugin to filter the returned plugins on.
        special_keys: Also return plugins which are under the special ('_') keys in this set.
        only_special_keys: Only return the plugins under the keys in ``special_keys`` and no others.

    Yields:
        Plugins in the ``PLUGINS`` tree based on the given filter criteria.
    """

    if osfilter is not None:
        # The PLUGINS tree does not include the hierarchy up to the plugins
        # directory (dissect.target.plugins) for the built-in plugins. For the
        # plugins in the directory specified in --plugin-path, the hierarchy
        # starts at that directory.
        #
        # Built-in OSPlugins do have the dissect.target.plugins path in their
        # module name, so it needs to be stripped, e.g.:
        #   dissect.target.plugins.general.default -> general.default
        #   dissect.target.plugins.windows._os -> plugins.windows._os
        #
        # The module name of OSPlugins from --plugin-path starts at the
        # directory specified in that option, e.g.:
        # --plugin-path=/some/path/, with a file foo/baros/_os.py
        # will have a module name of: foo.baros._os
        filter_path = _modulepath(osfilter).split(".")

        # If an OSPlugin is not defined in a file called _os.py, an extra `_os`
        # part is added to the PLUGINS tree.
        # For example the default OS plugin with module name general.default
        # (after stripping of the build-in hierarchy) will be added at:
        # general
        #        \- default
        #                  \- _os
        # However the `_os` part is not in the module name. Modules that are
        # defined in an _os.py file have the `_os` part in their module name.
        # It is stripped out, so the filter is similar for both types of
        # OSPlugin files.
        if filter_path[-1] == "_os":
            filter_path = filter_path[:-1]
    else:
        filter_path = []

    def _walk(
        root: dict,
        special_keys: set[str] = set(),
        only_special_keys: bool = False,
        prev_module_path: list[str] = [],
    ):
        for key, obj in root.items():
            module_path = prev_module_path.copy()
            module_path.append(key)

            # A branch in the PLUGINS tree is traversed to the next level if:
            # - there are no filters (which in effect means all plugins are
            # returned including all _os plugins).
            # - the osfilter is the default plugin (which means all normal plugins but
            # only the default _os plugin is returned).
            # - there is no _os plugin on the next level (we're traversing a
            # "normal" plugin branch or already jumped into an OS specific
            # branch because of a filter_path match)
            # - the current module_path fully matches the (beginning of) the
            # filter path (this allows traversing into the specific os branch
            # for the given os filter and any sub branches which are not os
            # branches (of a sub-os) themselves).
            if (
                not filter_path
                or issubclass(osfilter, general.default.DefaultPlugin)
                or "_os" not in obj
                or module_path == filter_path[: len(module_path)]
            ):
                if key.startswith("_"):
                    if key in special_keys:
                        # OSPlugins are treated special and are only returned
                        # if their module_path matches the full filter_path.
                        #
                        # Note that the module_path includes the `_os` part,
                        # which may have been explicitly added in the
                        # hierarchy. This part needs to be stripped out when
                        # matching against the filter_path, where it was either
                        # not present or stripped out.
                        if key != "_os" or (
                            key == "_os" and (not filter_path or (filter_path and module_path[:-1] == filter_path))
                        ):
                            # If the special key is a leaf-node, we just give it back.
                            # If it is a branch, we give back the full branch,
                            # not just the special_keys if only_special_keys
                            # was set to True.
                            if "functions" in obj:
                                yield obj
                            else:
                                yield from _walk(
                                    obj,
                                    special_keys=special_keys,
                                    only_special_keys=False,
                                    prev_module_path=module_path,
                                )
                        else:
                            continue
                    else:
                        continue

                else:
                    if "functions" in obj:
                        if not (special_keys and only_special_keys):
                            yield obj
                    else:
                        yield from _walk(
                            obj,
                            special_keys=special_keys,
                            only_special_keys=only_special_keys,
                            prev_module_path=module_path,
                        )

    yield from sorted(
        _walk(
            _get_plugins(),
            special_keys=special_keys,
            only_special_keys=only_special_keys,
        ),
        key=lambda plugin: len(plugin["module"]),
        reverse=True,
    )


def os_plugins() -> Iterator[PluginDescriptor]:
    """Retrieve all OS plugin descriptors."""
    yield from plugins(special_keys={"_os"}, only_special_keys=True)


def child_plugins() -> Iterator[PluginDescriptor]:
    """Retrieve all child plugin descriptors."""
    yield from plugins(special_keys={"_child"}, only_special_keys=True)


def lookup(func_name: str, osfilter: Optional[type[OSPlugin]] = None) -> Iterator[PluginDescriptor]:
    """Lookup a plugin descriptor by function name.

    Args:
        func_name: Function name to lookup.
        osfilter: The ``OSPlugin`` to use as template to find os specific plugins for.
    """
    yield from get_plugins_by_func_name(func_name, osfilter=osfilter)
    yield from get_plugins_by_namespace(func_name, osfilter=osfilter)


def get_plugins_by_func_name(func_name: str, osfilter: Optional[type[OSPlugin]] = None) -> Iterator[PluginDescriptor]:
    """Get a plugin descriptor by function name.

    Args:
        func_name: Function name to lookup.
        osfilter: The ``OSPlugin`` to use as template to find os specific plugins for.
    """
    for plugin_desc in plugins(osfilter):
        if not plugin_desc["namespace"] and func_name in plugin_desc["functions"]:
            yield plugin_desc


def get_plugins_by_namespace(namespace: str, osfilter: Optional[type[OSPlugin]] = None) -> Iterator[PluginDescriptor]:
    """Get a plugin descriptor by namespace.

    Args:
        namespace: Plugin namespace to match.
        osfilter: The ``OSPlugin`` to use as template to find os specific plugins for.
    """
    for plugin_desc in plugins(osfilter):
        if namespace == plugin_desc["namespace"]:
            yield plugin_desc


def load(plugin_desc: PluginDescriptor) -> Type[Plugin]:
    """Helper function that loads a plugin from a given plugin description.

    Args:
        plugin_desc: Plugin description as returned by plugin.lookup().

    Returns:
        The plugin class.

    Raises:
        PluginError: Raised when any other exception occurs while trying to load the plugin.
    """
    module = plugin_desc["fullname"].rsplit(".", 1)[0]

    try:
        module = importlib.import_module(module)
        return getattr(module, plugin_desc["class"])
    except Exception as e:
        raise PluginError(f"An exception occurred while trying to load a plugin: {module}") from e


def failed() -> list[dict[str, Any]]:
    """Return all plugins that failed to load."""
    return _get_plugins().get("_failed", [])


def _get_plugins() -> dict[str, PluginDescriptor]:
    """Load the plugin registry, or generate it if it doesn't exist yet."""
    global PLUGINS, GENERATED
    if not GENERATED:
        PLUGINS = generate()
        GENERATED = True
    return PLUGINS


def save_plugin_import_failure(module: str) -> None:
    """Store errors that occurred during plugin import."""
    if "_failed" not in PLUGINS:
        PLUGINS["_failed"] = []

    stacktrace = traceback.format_exception(*sys.exc_info())
    PLUGINS["_failed"].append(
        {
            "module": module,
            "stacktrace": stacktrace,
        }
    )


def find_py_files(plugin_path: Path) -> Iterator[Path]:
    """Walk all the files and directories in ``plugin_path`` and return all files ending in ``.py``.

    Do not walk or yield paths containing the following names:

    - __pycache__
    - __init__

    Furthermore, it logs an error if ``plugin_path`` does not exist.

    Args:
        plugin_path: The path to a directory or file to walk and filter.
    """
    if not plugin_path.exists():
        log.error("Path %s does not exist.", plugin_path)
        return

    if plugin_path.is_file():
        path_iterator = [plugin_path]
    else:
        path_iterator = plugin_path.glob("**/*.py")

    for path in path_iterator:
        if not path.is_file() or str(path).endswith("__init__.py"):
            continue

        yield path


def load_module_from_name(module_path: str) -> None:
    """Load a module from ``module_path``."""
    try:
        # This will trigger the __init__subclass__() of the Plugin subclasses in the module.
        importlib.import_module(module_path)
    except Exception as e:
        log.info("Unable to import %s", module_path)
        log.debug("Error while trying to import module %s", module_path, exc_info=e)
        save_plugin_import_failure(module_path)


def generate() -> dict[str, Any]:
    """Internal function to generate the list of available plugins.

    Walks the plugins directory and imports any .py files in there.
    Plugins will be automatically registered due to the decorators on them.

    Returns:
        The global ``PLUGINS`` dictionary.
    """
    global PLUGINS

    if "_failed" not in PLUGINS:
        PLUGINS["_failed"] = []

    plugins_dir = Path(__file__).parent / "plugins"
    for path in find_py_files(plugins_dir):
        relative_path = path.relative_to(plugins_dir)
        module_tuple = (MODULE_PATH, *relative_path.parent.parts, relative_path.stem)
        load_module_from_name(".".join(module_tuple))

    return PLUGINS


def load_module_from_file(path: Path, base_path: Path):
    """Loads a module from a file indicated by ``path`` relative to ``base_path``.

    The module is added to ``sys.modules`` so it can be found everywhere.

    Args:
        path: The file to load as module.
        base_path: The base directory of the module.
    """
    try:
        relative_path = path.relative_to(base_path)
        module_tuple = (*relative_path.parent.parts, relative_path.stem)
        spec = importlib.util.spec_from_file_location(".".join(module_tuple), path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        sys.modules[module.__name__] = module
    except Exception as e:
        log.error("Unable to import %s", path)
        log.debug("Error while trying to import module %s", path, exc_info=e)
        save_plugin_import_failure(str(path))


def load_modules_from_paths(plugin_dirs: list[Path]) -> None:
    """Iterate over the ``plugin_dirs`` and load all ``.py`` files."""
    for plugin_path in plugin_dirs:
        for path in find_py_files(plugin_path):
            base_path = plugin_path.parent if path == plugin_path else plugin_path
            load_module_from_file(path, base_path)


def get_external_module_paths(path_list: list[Path]) -> list[Path]:
    """Create a deduplicated list of paths."""
    output_list = environment_variable_paths() + path_list

    return list(set(output_list))


def environment_variable_paths() -> list[Path]:
    env_var = os.environ.get("DISSECT_PLUGINS")

    plugin_dirs = env_var.split(":") if env_var else []

    return [Path(directory) for directory in plugin_dirs]


def _traverse(key: str, obj: dict[str, Any]) -> dict[str, Any]:
    """Split a module path up in a dictionary."""
    for p in key.split("."):
        if p not in obj:
            obj[p] = {}

        obj = obj[p]

    return obj


def _modulepath(cls) -> str:
    """Returns the module path of a :class:`Plugin` relative to ``dissect.target.plugins``."""
    module = getattr(cls, "__module__", "")
    return module.replace(MODULE_PATH, "").lstrip(".")


# These need to be at the bottom of the module because __init_subclass__ requires everything
# in the parent class Plugin to be defined and resolved.
class NamespacePlugin(Plugin):
    def __init__(self, target: Target):
        """A namespace plugin provides services to access functionality from a group of subplugins.

        Support is currently limited to shared exported functions that yield records.
        """
        super().__init__(target)

        # The code below only applies to the direct subclass, indirect subclasses are finished here.
        if self.__class__ != self.__nsplugin__:
            return

        self._subplugins = []
        for entry in self.SUBPLUGINS:
            try:
                subplugin = getattr(self.target, entry)
                self._subplugins.append(subplugin)
            except UnsupportedPluginError:
                target.log.warning("Subplugin %s is not compatible with target.", entry)
            except Exception:
                target.log.exception("Failed to load subplugin: %s", entry)

    def check_compatible(self) -> None:
        if not len(self._subplugins):
            raise UnsupportedPluginError("No compatible subplugins found")

    def __init_subclass_namespace__(cls, **kwargs):
        # If this is a direct subclass of a Namespace plugin, create a reference to the current class for indirect
        # subclasses. This is necessary to autogenerate aggregate methods there
        cls.__nsplugin__ = cls
        cls.__findable__ = False

    def __init_subclass_subplugin__(cls, **kwargs):
        cls.__findable__ = True

        if not getattr(cls.__nsplugin__, "SUBPLUGINS", None):
            cls.__nsplugin__.SUBPLUGINS = set()

        # Register the current plugin class as a subplugin with
        # the direct subclass of NamespacePlugin
        cls.__nsplugin__.SUBPLUGINS.add(cls.__namespace__)

        # Generate a tuple of class names for which we do not want to add subplugin functions, which is the
        # namespaceplugin and all of its superclasses (minus the base object).
        reserved_cls_names = tuple({_class.__name__ for _class in cls.__nsplugin__.mro() if _class is not object})

        # Collect the public attrs of the subplugin
        for subplugin_func_name in cls.__exports__:
            subplugin_func = inspect.getattr_static(cls, subplugin_func_name)

            # The attr need to be callable and exported
            if not isinstance(subplugin_func, Callable):
                continue

            # The method needs to output records
            if getattr(subplugin_func, "__output__", None) not in ["record", "yield"]:
                continue

            # The method may not be part of a parent class.
            if subplugin_func.__qualname__.startswith(reserved_cls_names):
                continue

            # If we already have an aggregate method, skip
            if existing_aggregator := getattr(cls.__nsplugin__, subplugin_func_name, None):
                if not hasattr(existing_aggregator, "__subplugins__"):
                    # This is not an aggregator, but a re-implementation of a subclass function by the subplugin.
                    continue
                existing_aggregator.__subplugins__.append(cls.__namespace__)
                continue

            # The generic template for the aggregator method
            def generate_aggregator(method_name: str) -> Callable:
                def aggregator(self) -> Iterator[Record]:
                    for entry in aggregator.__subplugins__:
                        try:
                            subplugin = getattr(self.target, entry)
                            yield from getattr(subplugin, method_name)()
                        except UnsupportedPluginError:
                            continue
                        except Exception as e:
                            self.target.log.error("Subplugin: %s raised an exception for: %s", entry, method_name)
                            self.target.log.debug("Exception: %s", e, exc_info=e)

                # Holds the subplugins that share this method
                aggregator.__subplugins__ = []

                return aggregator

            # The generic template for the documentation method
            def generate_documentor(cls, method_name: str, aggregator: Callable) -> str:
                def documentor():
                    return defaultdict(
                        lambda: "???",
                        {
                            "func_name": f"{cls.__nsplugin__.__namespace__}.{method_name}",
                            "short_description": "".join(
                                [
                                    f"Return {method_name} for: ",
                                    ",".join(aggregator.__subplugins__),
                                ]
                            ),
                            "output_type": "records",
                        },
                    )

                return documentor

            # Manifacture a method for the namespaced class
            generated_aggregator = generate_aggregator(subplugin_func_name)
            generated_documentor = generate_documentor(cls, subplugin_func_name, generated_aggregator)

            # Add as an attribute to the namespace class
            setattr(cls.__nsplugin__, subplugin_func_name, generated_aggregator)

            # Copy the meta descriptors of the function attribute
            for copy_attr in ["__output__", "__record__", "__doc__", "__exported__"]:
                setattr(generated_aggregator, copy_attr, getattr(subplugin_func, copy_attr, None))

            # Add subplugin to aggregator
            generated_aggregator.__subplugins__.append(cls.__namespace__)

            # Mark the function as being autogenerated
            setattr(generated_aggregator, "__autogen__", True)

            # Add the documentor function to the aggregator
            setattr(generated_aggregator, "get_func_doc_spec", generated_documentor)

            # Register the newly auto-created method
            cls.__nsplugin__.__exports__.append(subplugin_func_name)
            cls.__nsplugin__.__functions__.append(subplugin_func_name)

    def __init_subclass__(cls, **kwargs):
        # Upon subclassing, decide whether this is a direct subclass of NamespacePlugin
        # If this is not the case, autogenerate aggregate methods for methods record output.
        super().__init_subclass__(**kwargs)
        if cls.__bases__[0] != NamespacePlugin:
            cls.__init_subclass_subplugin__(cls, **kwargs)
        else:
            cls.__init_subclass_namespace__(cls, **kwargs)


__COMMON_PLUGIN_METHOD_NAMES__ = {attr.__name__ for attr in get_nonprivate_methods(Plugin)}


class InternalPlugin(Plugin):
    """Parent class for internal plugins.

    InternalPlugin marks all non-private methods internal by default
    (same as ``@internal`` decorator).
    """

    def __init_subclass__(cls, **kwargs):
        for method in get_nonprivate_methods(cls):
            if callable(method) and method.__name__ not in __COMMON_PLUGIN_METHOD_NAMES__:
                method.__internal__ = True

        super().__init_subclass__(**kwargs)
        return cls


class InternalNamespacePlugin(NamespacePlugin, InternalPlugin):
    pass


@dataclass(frozen=True, eq=True)
class PluginFunction:
    name: str
    path: str
    output_type: str
    class_object: type[Plugin]
    method_name: str
    plugin_desc: PluginDescriptor = field(hash=False)

    def __repr__(self) -> str:
        return self.path


def plugin_function_index(target: Optional[Target]) -> tuple[dict[str, PluginDescriptor], set[str]]:
    """Returns an index-list for plugins.

    This list is used to match CLI expressions against to find the desired plugin.
    Also returns the roots to determine whether a CLI expression has to be compared
    to the plugin tree or parsed using legacy rules.
    """

    if target is None:
        os_type = None
    elif target._os_plugin is None:
        os_type = general.default.DefaultPlugin
    elif isinstance(target._os_plugin, type) and issubclass(target._os_plugin, OSPlugin):
        os_type = target._os_plugin
    elif isinstance(target._os_plugin, OSPlugin):
        os_type = type(target._os_plugin)
    else:
        raise TypeError(
            "target must be None or target._os_plugin must be either None, "
            "a subclass of OSPlugin or an instance of OSPlugin"
        )

    index = {}
    rootset = set()

    all_plugins = plugins(osfilter=os_type, special_keys={"_child", "_os"})

    for available_original in all_plugins:
        # Prevent modifying the global PLUGINS dict, otherwise -f os.windows._os.users fails for instance.
        available = available_original.copy()

        modulepath = available["module"]
        rootset.add(modulepath.split(".")[0])

        if "get_all_records" in available["exports"]:
            # The get_all_records does not only need to be not present in the
            # index, it also needs to be removed from the exports list, else
            # the 'plugins' plugin will still display them.
            exports = available["exports"].copy()
            exports.remove("get_all_records")
            available["exports"] = exports

            if "get_all_records" not in available_original["exports"]:
                raise Exception(f"get_all_records removed from {available_original}")

        for exported in available["exports"]:
            if available["is_osplugin"] and os_type == general.default.DefaultPlugin:
                # This makes the os plugin exports listed under the special
                # "OS plugins" header by the 'plugins' plugin.
                available["module"] = ""

            index[f"{modulepath}.{exported}"] = available

    return index, rootset


def find_plugin_functions(
    target: Optional[Target],
    patterns: str,
    compatibility: bool = False,
    **kwargs,
) -> tuple[list[PluginFunction], set[str]]:
    """Finds plugins that match the target and the patterns.

    Given a target, a comma separated list of patterns and an optional compatibility flag,
    this function finds matching plugins, optionally checking compatibility and returns
    a list of plugin function descriptors (including output types).
    """
    result = []

    functions, rootset = plugin_function_index(target)

    invalid_funcs = set()
    show_hidden = kwargs.get("show_hidden", False)
    ignore_load_errors = kwargs.get("ignore_load_errors", False)

    for pattern in patterns.split(","):
        # Backward compatibility fix for namespace-level plugins (i.e. chrome)
        # If an exact namespace match is found, the pattern is changed to the tree to that namespace.
        # Examples:
        #     -f browser -> apps.browser.browser
        #     -f iexplore -> apps.browser.iexplore
        namespace_match = False
        for index_name, func in functions.items():
            if func["namespace"] == pattern:
                pattern = func["module"]
                namespace_match = True
                break

        wildcard = any(char in pattern for char in ["*", "!", "?", "[", "]"])
        treematch = pattern.split(".")[0] in rootset and pattern != "os"
        exact_match = pattern in functions

        # Allow for exact and namespace matches even if the plugin does not want to be found, otherwise you cannot
        # reach documented namespace plugins like apps.browser.browser.downloads.
        # You can *always* run these using the namespace/classic-style like: browser.downloads (but -l lists them
        # in the tree for documentation purposes so it would be misleading not to allow tree access as well).
        #
        # Note that these tree items will never respond to wildcards though to avoid duplicate results, e.g. when
        # querying apps.browser.*, this also means apps.browser.browser.* won't work.
        if exact_match or namespace_match:
            show_hidden = True

        # Change the treematch pattern into an fnmatch-able pattern to give back all functions from the sub-tree
        # (if there is a subtree).
        #
        # Examples:
        #     -f browser -> apps.browser.browser* (the whole package, due to a namespace match)
        #     -f apps.webservers.iis -> apps.webservers.iis* (logs etc)
        #     -f apps.webservers.iis.logs -> apps.webservers.iis.logs* (only the logs, there is no subtree)
        # We do not include a dot because that does not work if the full path is given:
        #     -f apps.webservers.iis.logs != apps.webservers.iis.logs.* (does not work)
        #
        # In practice a namespace_match would almost always also be a treematch, except when the namespace plugin
        # is in the root of the plugin tree.
        if (treematch or namespace_match) and not wildcard and not exact_match:
            pattern += "*"

        if wildcard or treematch:
            matches = False
            for index_name in fnmatch.filter(functions.keys(), pattern):
                func = functions[index_name]

                method_name = index_name.split(".")[-1]
                try:
                    loaded_plugin_object = load(func)
                except Exception:
                    if ignore_load_errors:
                        continue
                    raise

                # Skip plugins that don't want to be found by wildcards
                if not show_hidden and not loaded_plugin_object.__findable__:
                    continue

                fobject = inspect.getattr_static(loaded_plugin_object, method_name)

                if compatibility:
                    if target is None:
                        continue
                    try:
                        if not loaded_plugin_object(target).is_compatible():
                            continue
                    except Exception:
                        continue

                matches = True
                result.append(
                    PluginFunction(
                        name=f"{func['namespace']}.{method_name}" if func["namespace"] else method_name,
                        path=index_name,
                        class_object=loaded_plugin_object,
                        method_name=method_name,
                        output_type=getattr(fobject, "__output__", "none"),
                        plugin_desc=func,
                    )
                )

            if not matches:
                invalid_funcs.add(pattern)

        else:
            # otherwise match using ~ classic style
            if pattern.find(".") > -1:
                namespace, funcname = pattern.split(".", 1)
            else:
                funcname = pattern
                namespace = None

            plugin_descriptions = []
            for func_path, func in functions.items():
                nsmatch = namespace and func["namespace"] == namespace and func_path.split(".")[-1] == funcname
                fmatch = not namespace and not func["namespace"] and func_path.split(".")[-1] == funcname
                if nsmatch or fmatch:
                    plugin_descriptions.append(func)

            if not plugin_descriptions:
                invalid_funcs.add(pattern)

            for description in plugin_descriptions:
                try:
                    loaded_plugin_object = load(description)
                except Exception:
                    if ignore_load_errors:
                        continue
                    raise

                fobject = inspect.getattr_static(loaded_plugin_object, funcname)

                if compatibility and not loaded_plugin_object(target).is_compatible():
                    continue

                result.append(
                    PluginFunction(
                        name=f"{description['namespace']}.{funcname}" if description["namespace"] else funcname,
                        path=f"{description['module']}.{funcname}",
                        class_object=loaded_plugin_object,
                        method_name=funcname,
                        output_type=getattr(fobject, "__output__", "none"),
                        plugin_desc=description,
                    )
                )

    return result, invalid_funcs



from typing import Iterator

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin


class VirtuozzoChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields from Virtuozzo container's root.

    Virtuozzo conatiners are by default registered in the folder ``vz/root/$VEID``,
    where VEID will be substituted with the actual container UUID.

    .. code-block::

        /
        etc/
        var/
        vz/
            root/
                <container-uuid>/
                <container-uuid>/

    References:
        - https://docs.virtuozzo.com/virtuozzo_hybrid_server_7_command_line_reference/managing-system/configuration-files.html
    """  # noqa: E501

    __type__ = "virtuozzo"

    PATH = "/vz/root"

    def check_compatible(self) -> None:
        if not self.target.fs.path(self.PATH).exists():
            raise UnsupportedPluginError("No Virtuozzo path found")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for container in self.target.fs.path(self.PATH).iterdir():
            yield ChildTargetRecord(
                type=self.__type__,
                path=container,
                _target=self.target,
            )



from __future__ import annotations

import io
import itertools
from itertools import product
from typing import Iterable, Iterator, Optional, Union

from dissect.target.exceptions import ConfigurationParsingError, UnsupportedPluginError
from dissect.target.helpers import fsutil
from dissect.target.helpers.configutil import Default, ListUnwrapper, _update_dictionary
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import OperatingSystem, Plugin, arg, export

COMMON_ELEMENTS = [
    ("string", "name"),  # basename of .conf file
    ("string", "proto"),
    ("string", "dev"),
    ("string", "ca"),
    ("string", "cert"),
    ("string", "key"),
    ("boolean", "redacted_key"),
    ("string", "tls_auth"),
    ("string", "status"),
    ("string", "log"),
    ("string", "source"),
]

OpenVPNServer = TargetRecordDescriptor(
    "application/vpn/openvpn/server",
    [
        ("net.ipaddress", "local"),
        ("uint16", "port"),
        ("string", "dh"),
        ("string", "topology"),
        ("string", "server"),
        ("string", "ifconfig_pool_persist"),
        ("string[]", "pushed_options"),
        ("boolean", "client_to_client"),
        ("boolean", "duplicate_cn"),
        *COMMON_ELEMENTS,
    ],
)

OpenVPNClient = TargetRecordDescriptor(
    "application/vpn/openvpn/client",
    [
        ("string[]", "remote"),
        *COMMON_ELEMENTS,
    ],
)


class OpenVPNParser(Default):
    def __init__(self, *args, **kwargs):
        boolean_fields = OpenVPNServer.getfields("boolean") + OpenVPNClient.getfields("boolean")
        self.boolean_field_names = set(field.name.replace("_", "-") for field in boolean_fields)

        super().__init__(*args, separator=(r"\s",), collapse=["key", "ca", "cert"], **kwargs)

    def parse_file(self, fh: io.TextIOBase) -> None:
        root = {}
        iterator = self.line_reader(fh)
        for line in iterator:
            if line.startswith("<"):
                key = line.strip().strip("<>")
                value = self._read_blob(iterator)
                _update_dictionary(root, key, value)
                continue

            self._parse_line(root, line)

        self.parsed_data = ListUnwrapper.unwrap(root)

    def _read_blob(self, lines: Iterable[str]) -> str | list[dict]:
        """Read the whole section between <data></data> sections"""
        output = ""
        with io.StringIO() as buffer:
            for line in lines:
                if "</" in line:
                    break

                buffer.write(line)
            output = buffer.getvalue()

        # Check for connection profile blocks
        if not output.startswith("-----"):
            profile_dict = dict()
            for line in output.splitlines():
                self._parse_line(profile_dict, line)

            # We put it as a list as _update_dictionary appends data in a list.
            output = [profile_dict]

        return output

    def _parse_line(self, root: dict, line: str) -> None:
        key, *value = self.SEPARATOR.split(line, 1)
        # Unquote data
        value = value[0].strip() if value else ""

        value = value.strip("'\"")

        if key in self.boolean_field_names:
            value = True

        _update_dictionary(root, key, value)


class OpenVPNPlugin(Plugin):
    """OpenVPN configuration parser.

    References:
        - man (8) openvpn
    """

    __namespace__ = "openvpn"

    config_globs = [
        # This catches openvpn@, openvpn-client@, and openvpn-server@ systemd configurations
        # Linux
        "/etc/openvpn/",
        # Windows
        "sysvol/Program Files/OpenVPN/config/",
    ]

    user_config_paths = {
        OperatingSystem.WINDOWS.value: ["OpenVPN/config/"],
        OperatingSystem.OSX.value: ["Library/Application Support/OpenVPN Connect/profiles/"],
    }

    def __init__(self, target) -> None:
        super().__init__(target)
        self.configs: list[fsutil.TargetPath] = []
        for base, glob in product(self.config_globs, ["*.conf", "*.ovpn"]):
            self.configs.extend(self.target.fs.path(base).rglob(glob))

        user_paths = self.user_config_paths.get(target.os, [])
        for path, glob, user_details in itertools.product(
            user_paths, ["*.conf", "*.ovpn"], self.target.user_details.all_with_home()
        ):
            self.configs.extend(user_details.home_path.joinpath(path).rglob(glob))

    def check_compatible(self) -> None:
        if not self.configs:
            raise UnsupportedPluginError("No OpenVPN configuration files found")

    def _load_config(self, parser: OpenVPNParser, config_path: fsutil.TargetPath) -> Optional[dict]:
        with config_path.open("rt") as file:
            try:
                parser.parse_file(file)
            except ConfigurationParsingError as e:
                # Couldn't parse file, continue
                self.target.log.info("An issue occurred during parsing of %s, continuing", config_path)
                self.target.log.debug("", exc_info=e)
                return None

        return parser.parsed_data

    @export(record=[OpenVPNServer, OpenVPNClient])
    @arg("--export-key", action="store_true")
    def config(self, export_key: bool = False) -> Iterator[Union[OpenVPNServer, OpenVPNClient]]:
        """Parses config files from openvpn interfaces."""
        # We define the parser here so we can reuse it
        parser = OpenVPNParser()

        for config_path in self.configs:
            config = self._load_config(parser, config_path)

            common_elements = {
                "name": config_path.stem,
                "proto": config.get("proto", "udp"),  # Default is UDP
                "dev": config.get("dev"),
                "ca": config.get("ca"),
                "cert": config.get("cert"),
                "key": config.get("key"),
                "status": config.get("status"),
                "log": config.get("log"),
                "source": config_path,
                "_target": self.target,
            }

            if not export_key and "PRIVATE KEY" in common_elements.get("key"):
                common_elements.update({"key": None})
                common_elements.update({"redacted_key": True})

            tls_auth = config.get("tls-auth", "")
            # The format of tls-auth is 'tls-auth ta.key <NUM>'.
            # NUM is either 0 or 1 depending on whether the configuration
            # is for the client or server, and that does not interest us
            # This gets rid of the number at the end, while still supporting spaces
            tls_auth = " ".join(tls_auth.split(" ")[:-1]).strip("'\"")

            common_elements.update({"tls_auth": tls_auth})

            if "client" in config:
                remote = config.get("remote", [])

                yield OpenVPNClient(
                    **common_elements,
                    remote=remote,
                )
            else:
                # Defaults here are taken from `man (8) openvpn`
                yield OpenVPNServer(
                    **common_elements,
                    local=config.get("local", "0.0.0.0"),
                    port=int(config.get("port", "1194")),
                    dh=config.get("dh"),
                    topology=config.get("topology"),
                    server=config.get("server"),
                    ifconfig_pool_persist=config.get("ifconfig-pool-persist"),
                    pushed_options=config.get("push", []),
                    client_to_client=config.get("client-to-client", False),
                    duplicate_cn=config.get("duplicate-cn", False),
                )



from unittest.mock import call, patch

from dissect.target.loaders.pvs import PvsLoader


@patch("pathlib.Path")
@patch("dissect.target.loaders.pvs.HddContainer")
@patch("dissect.target.loaders.pvs.pvs.PVS")
def test_pvs_loader(PVS, HddContainer, Path, target_bare):
    PVS.return_value = PVS
    PVS.disks.return_value = ["mock.hdd"]
    HddContainer.return_value = HddContainer

    pvs_loader = PvsLoader(Path("/mock.pvs"))
    pvs_loader.map(target_bare)

    expected = [call.disks()]
    del PVS.mock_calls[0]
    assert expected == PVS.mock_calls
    assert len(target_bare.disks) == 1



from __future__ import annotations

from datetime import datetime, timezone
from io import BytesIO
from pathlib import Path
from unittest.mock import patch
from zoneinfo import ZoneInfo

import pytest
from flow.record.fieldtypes import datetime as dt

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.log.auth import AuthPlugin
from dissect.target.target import Target
from tests._utils import absolute_path


def test_auth_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO("Europe/Amsterdam".encode()))

    data_path = "_data/plugins/os/unix/log/auth/auth.log"
    data_file = absolute_path(data_path)
    fs_unix.map_file("var/log/auth.log", data_file)

    entry = fs_unix.get("var/log/auth.log")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix.add_plugin(AuthPlugin)
        results = list(target_unix.authlog())

        assert len(results) == 10
        assert results[-1].ts == dt(2022, 11, 14, 6, 39, 1, tzinfo=ZoneInfo("Europe/Amsterdam"))
        assert results[-1].message == "pam_unix(cron:session): session opened for user root by (uid=0)"


def test_auth_plugin_with_gz(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO("Pacific/Honolulu".encode()))

    empty_file = absolute_path("_data/plugins/os/unix/log/empty.log")
    fs_unix.map_file("var/log/auth.log", empty_file)

    gz_path = "_data/plugins/os/unix/log/auth/auth.log.gz"
    gz_file = absolute_path(gz_path)
    fs_unix.map_file("var/log/auth.log.gz", gz_file)

    entry = fs_unix.get("var/log/auth.log.gz")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix.add_plugin(AuthPlugin)
        results = list(target_unix.authlog())

        assert len(results) == 10
        assert results[-1].ts == dt(2022, 11, 14, 6, 39, 1, tzinfo=ZoneInfo("Pacific/Honolulu"))
        assert results[-1].message == "pam_unix(cron:session): session opened for user root by (uid=0)"


def test_auth_plugin_with_bz(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO("America/Nuuk".encode()))

    empty_file = absolute_path("_data/plugins/os/unix/log/empty.log")
    fs_unix.map_file("var/log/auth.log", empty_file)

    bz_path = "_data/plugins/os/unix/log/auth/auth.log.bz2"
    bz_file = absolute_path(bz_path)
    fs_unix.map_file("var/log/auth.log.1.bz2", bz_file)

    entry = fs_unix.get("var/log/auth.log.1.bz2")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix.add_plugin(AuthPlugin)
        results = list(target_unix.authlog())

        assert len(results) == 10
        assert results[-1].ts == dt(2022, 11, 14, 6, 39, 1, tzinfo=ZoneInfo("America/Nuuk"))
        assert results[-1].message == "pam_unix(cron:session): session opened for user root by (uid=0)"


def test_auth_plugin_year_rollover(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh("/etc/timezone", BytesIO("Etc/UTC".encode()))

    data_path = "_data/plugins/os/unix/log/auth/secure"
    data_file = absolute_path(data_path)
    fs_unix.map_file("var/log/secure", data_file)

    entry = fs_unix.get("var/log/secure")
    stat = entry.stat()
    stat.st_mtime = datetime(2022, 6, 1, tzinfo=timezone.utc).timestamp()

    with patch.object(entry, "stat", return_value=stat):
        target_unix.add_plugin(AuthPlugin)
        results = list(target_unix.authlog())

        assert len(results) == 2
        results.reverse()
        assert results[0].ts == dt(2021, 12, 31, 3, 14, 0, tzinfo=ZoneInfo("Etc/UTC"))
        assert results[1].ts == dt(2022, 1, 1, 13, 37, 0, tzinfo=ZoneInfo("Etc/UTC"))


@pytest.mark.parametrize(
    "message, results",
    [
        pytest.param(
            "Mar 29 10:43:01 ubuntu-1 sshd[1193]: Accepted password for test_user from 8.8.8.8 port 52942 ssh2",
            {
                "service": "sshd",
                "pid": 1193,
                "action": "accepted authentication",
                "authentication_type": "password",
                "user": "test_user",
                "remote_ip": "8.8.8.8",
                "port": 52942,
            },
            id="sshd: accepted password",
        ),
        pytest.param(
            "Jun 4 22:14:15 ubuntu-1 sshd[41458]: Failed password for root from 8.8.8.8 port 22 ssh2",
            {
                "service": "sshd",
                "pid": 41458,
                "action": "failed authentication",
                "authentication_type": "password",
                "user": "root",
                "remote_ip": "8.8.8.8",
                "port": 22,
            },
            id="sshd: failed password",
        ),
        pytest.param(
            "Mar 27 13:08:09 ubuntu-1 sshd[1361]: Accepted publickey for test_user "
            "from 8.8.8.8 port 12345 ssh2: RSA SHA256:123456789asdfghjklertzuio",
            {
                "service": "sshd",
                "pid": 1361,
                "action": "accepted authentication",
                "authentication_type": "publickey",
                "user": "test_user",
                "remote_ip": "8.8.8.8",
                "port": 12345,
                "ssh_protocol": "ssh2",
                "encryption_algorithm": "RSA",
                "hash_algorithm": "SHA256",
                "key_hash": "123456789asdfghjklertzuio",
            },
            id="sshd: accepted publickey",
        ),
        pytest.param(
            "Mar 27 13:08:09 ubuntu-1 sshd[1361]: Failed publickey for test_user from 8.8.8.8 port 12345 ssh2.",
            {
                "service": "sshd",
                "pid": 1361,
                "action": "failed authentication",
                "authentication_type": "publickey",
                "user": "test_user",
                "remote_ip": "8.8.8.8",
                "port": 12345,
            },
            id="sshd: failed publickey",
        ),
        pytest.param(
            "Mar 27 13:06:56 ubuntu-1 sshd[1291]: Server listening on 127.0.0.1 port 22.",
            {
                "service": "sshd",
                "pid": 1291,
                "host_ip": "127.0.0.1",
                "port": 22,
            },
            id="sshd: listening",
        ),
        pytest.param(
            "Mar 27 13:08:09 ubuntu-1 sshd[1361]: pam_unix(sshd:session): session opened for user test_user by (uid=0)",
            {
                "service": "sshd",
                "pid": 1361,
                "action": "session opened",
                "user": "test_user",
                "user_uid": None,
                "by_uid": 0,
            },
            id="sshd: pam_unix",
        ),
        pytest.param(
            "Mar 27 13:08:09 ubuntu-1 sshd[1361]: pam_unix(sshd:session): session opened "
            "for user root(uid=0) by (uid=0)",
            {
                "service": "sshd",
                "pid": 1361,
                "action": "session opened",
                "user": "root",
                "user_uid": 0,
                "by_uid": 0,
            },
            id="sshd: pam_unix",
        ),
        pytest.param(
            "Mar 27 13:06:56 ubuntu-1 systemd-logind[1118]: Watching system buttons "
            "on /dev/input/event0 (Power Button)",
            {
                "service": "systemd-logind",
                "pid": 1118,
                "action": "Watching system buttons",
                "device": "/dev/input/event0",
                "device_name": "Power Button",
            },
            id="systemd-logind: watching system buttons",
        ),
        pytest.param(
            "Mar 27 13:06:56 ubuntu-1 systemd-logind[1118]: New seat seat0.",
            {
                "service": "systemd-logind",
                "pid": 1118,
                "action": "new seat",
                "seat": "seat0",
            },
            id="systemd-logind: new seat",
        ),
        pytest.param(
            "Mar 27 13:10:08 ubuntu-1 sudo:   ubuntu : TTY=pts/0 ; PWD=/home/test_user ; "
            "USER=root ; COMMAND=/usr/bin/apt-key add -",
            {
                "service": "sudo",
                "pid": None,
                "tty": "pts/0",
                "pwd": "/home/test_user",
                "effective_user": "root",
                "command": "/usr/bin/apt-key add -",
            },
            id="sudo: command",
        ),
        pytest.param(
            "Apr  3 12:32:23 ubuntu-1 su[1521]: Successful su for user by root",
            {"service": "su", "pid": 1521, "su_result": "success", "user": "user", "by": "root"},
            id="su: success",
        ),
        pytest.param(
            "Apr  3 12:32:23 ubuntu-1 su[1531]: 'su root' failed for user by root",
            {
                "service": "su",
                "pid": 1531,
                "su_result": "failed",
                "command": "su root",
                "user": "user",
                "by": "root",
            },
            id="su: failed",
        ),
        pytest.param(
            "Apr  3 12:32:23 ubuntu-1 pkexec[1531]: user: Executing command [USER=root] "
            "[TTY=unknown] [CWD=/home/user] [COMMAND=/usr/lib/update-notifier/package-system-locked]",
            {
                "service": "pkexec",
                "pid": 1531,
                "action": "executing command",
                "user": "user",
                "effective_user": "root",
                "tty": "unknown",
                "cwd": "/home/user",
                "command": "/usr/lib/update-notifier/package-system-locked",
            },
            id="pkexec: executing command",
        ),
        pytest.param(
            "Mar 27 13:17:01 ubuntu-1 CRON[2623]: pam_unix(cron:session): session closed for user root",
            {
                "service": "CRON",
                "pid": 2623,
                "action": "session closed",
                "user": "root",
            },
            id="cron: pam_unix",
        ),
    ],
)
def test_auth_plugin_additional_fields(
    target_unix, fs_unix: VirtualFilesystem, tmp_path: Path, message: str, results: dict[str, str | int]
) -> None:
    data_path = tmp_path / "auth.log"
    data_path.write_text(message)
    fs_unix.map_file("var/log/auth.log", data_path)

    target_unix.add_plugin(AuthPlugin)
    record = list(target_unix.authlog())[0]

    for key, value in results.items():
        assert getattr(record, key) == value


def test_auth_plugin_iso_date_format(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """test if we correctly handle Ubuntu 24.04 ISO formatted dates."""

    fs_unix.map_file("/var/log/auth.log", absolute_path("_data/plugins/os/unix/log/auth/iso.log"))
    target_unix.add_plugin(AuthPlugin)

    results = sorted(list(target_unix.authlog()), key=lambda r: r.ts)
    assert len(results) == 10

    assert results[0].ts == datetime(2024, 12, 31, 11, 37, 1, 123456, tzinfo=timezone.utc)
    assert results[0].service == "sudo"
    assert results[0].pid is None
    assert results[0].tty == "pts/0"
    assert results[0].pwd == "/home/user"
    assert results[0].effective_user == "root"
    assert results[0].command == "/usr/bin/chmod go+r /etc/apt/keyrings/githubcli-archive-keyring.gpg"
    assert results[0].source == "/var/log/auth.log"
    assert (
        results[0].message
        == "user : TTY=pts/0 ; PWD=/home/user ; USER=root ; COMMAND=/usr/bin/chmod go+r /etc/apt/keyrings/githubcli-archive-keyring.gpg"  # noqa: E501
    )
